#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Start subsystems (API server, logging daemon, etc.).
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Optional, List, Any

def start(
        action: Optional[List[str]] = None,
        **kw: Any,
    ) -> SuccessTuple:
    """
    Start subsystems (API server, background job, etc.).
    """
    from meerschaum.actions import choose_subaction
    options = {
        'api': _start_api,
        'jobs': _start_jobs,
        'gui': _start_gui,
        'webterm': _start_webterm,
        'connectors': _start_connectors,
    }
    return choose_subaction(action, options, **kw)


def _complete_start(
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> List[str]:
    """
    Override the default Meerschaum `complete_` function.
    """

    if action is None:
        action = []

    options = {
        'job': _complete_start_jobs,
        'jobs': _complete_start_jobs,
        'connector': _complete_start_connectors,
        'connectors': _complete_start_connectors,
    }

    if (
        len(action) > 0 and action[0] in options
            and kw.get('line', '').split(' ')[-1] != action[0]
    ):
        sub = action[0]
        del action[0]
        return options[sub](action=action, **kw)

    from meerschaum._internal.shell import default_action_completer
    return default_action_completer(action=(['start'] + action), **kw)


def _start_api(action: Optional[List[str]] = None, **kw):
    """
    Start the API server.
    
    Usage:
        `start api {options}`
    
    Options:
        - `-p, --port {number}`
            Port to bind the API server to.
    
        - `-w, --workers {number}`
            How many worker threads to run.
            Defaults to the number of CPU cores or 1 on Android.
    """
    from meerschaum.actions import actions
    return actions['api'](action=['start'], **kw)

def _start_jobs(
        action: Optional[List[str]] = None,
        name: Optional[str] = None,
        **kw
    ) -> SuccessTuple:
    """
    Run a Meerschaum action as a background job.
    
    To create a new job, pass the command arguments after `start job`.
    To start a stopped job, pass the job name after `start job`.
    
    You may also run a background job with the `-d` or `--daemon` flags.
    
    Examples:
    
        Create new jobs:
    
          - `start job sync pipes --loop`
                Run the action `sync pipes --loop` as a background job.
                Generates a random name; e.g. 'happy_seal'.
    
          - `start api --daemon --name api_server`
                Run the action `start api` as a background job, and assign the job
                the name 'api_server'.
    
        Start stopped jobs:
    
          - `start job happy_seal`
                Start the job 'happy_seal'.
    
          - `start job --name happy_seal`
                Start the job 'happy_seal' but via the `--name` flag.
                This only applies when no text follows the words 'start job'.
    """
    import textwrap
    from meerschaum.utils.warnings import warn, info
    from meerschaum.utils.daemon import (
        daemon_action, Daemon, get_daemon_ids, get_daemons, get_filtered_daemons,
        get_stopped_daemons, get_running_daemons, get_paused_daemons,
    )
    from meerschaum.utils.daemon._names import get_new_daemon_name
    from meerschaum._internal.arguments._parse_arguments import parse_arguments
    from meerschaum.actions import actions
    from meerschaum.utils.prompt import yes_no
    from meerschaum.utils.formatting import print_tuple
    from meerschaum.utils.formatting._jobs import pprint_job, pprint_jobs
    from meerschaum.utils.formatting._shell import clear_screen
    from meerschaum.utils.misc import items_str

    names = []
    daemon_ids = get_daemon_ids()

    new_job = len(list(action)) > 0
    _potential_jobs = {'known': [], 'unknown': []}

    if action:
        for a in action:
            _potential_jobs[(
                'known'
                if a in daemon_ids
                else 'unknown'
            )].append(a)

        ### Check if the job is named after an action.
        if (
            _potential_jobs['known']
                and _potential_jobs['unknown']
                and _potential_jobs['known'][0] == action[0]
                and _potential_jobs['known'][0] in actions
        ):
            _potential_jobs['unknown'].insert(0, _potential_jobs['known'][0])
            del _potential_jobs['known'][0]

        ### Only spawn a new job if we don't find any jobs.
        new_job = (len(_potential_jobs['known']) == 0)
        if not new_job and _potential_jobs['unknown']:
            if not kw.get('nopretty', False):
                warn(
                    (
                        "Unknown job" + ("s" if len(_potential_jobs['unknown']) > 1 else '') + " "
                        + items_str(_potential_jobs['unknown'])
                        + " will be ignored."
                    ),
                    stack = False
                )

        ### Determine the `names` list.
        if new_job:
            names = [(get_new_daemon_name() if not name else name)]
        elif not new_job and not name:
            names = _potential_jobs['known']
        ### Cannot find dameon_id
        else:
            msg = (
                f"Unknown job" + ('s' if len(action) != 1 else '') + ' '
                + items_str(action, and_str='or') + '.'
            )
            return False, msg

    ### No action provided but a --name was. Start job if possible.
    ### E.g. `start job --name myjob`
    elif name is not None:
        new_job = False
        names = [name]

    ### No action or --name was provided. Ask to start all stopped jobs.
    else:
        _running_daemons = get_running_daemons()
        _paused_daemons = get_paused_daemons()
        _stopped_daemons = get_stopped_daemons()
        if not _stopped_daemons and not _paused_daemons:
            if not _running_daemons:
                return False, "No jobs to start."
            return True, "All jobs are running."

        names = [d.daemon_id for d in _stopped_daemons + _paused_daemons]

    def _run_new_job(name: Optional[str] = None):
        kw['action'] = action
        name = name or get_new_daemon_name()
        kw['name'] = name
        _action_success_tuple = daemon_action(daemon_id=name, **kw)
        return _action_success_tuple, name

    def _run_existing_job(name: Optional[str] = None):
        daemon = Daemon(daemon_id=name)
        if daemon.process is not None:
            if daemon.status == 'paused':
                return daemon.resume(), daemon.daemon_id
            return (True, f"Job '{name}' is already running."), daemon.daemon_id

        if not daemon.path.exists():
            if not kw.get('nopretty', False):
                warn(f"There isn't a job with the name '{name}'.", stack=False)
                print(
                    f"You can start a new job named '{name}' with `start job "
                    + "{options}" + f" --name {name}`"
                )
            return (False, f"Job '{name}' does not exist."), daemon.daemon_id

        return daemon.run(allow_dirty_run=True), daemon.daemon_id

    if not names:
        return False, "No jobs to start."

    ### Get user permission to clear logs.
    _filtered_daemons = get_filtered_daemons(names)
    if not kw.get('force', False) and _filtered_daemons:
        _filtered_running_daemons = get_running_daemons(_filtered_daemons)
        _skipped_daemons = []
        if _filtered_running_daemons:
            pprint_jobs(_filtered_running_daemons)
            if yes_no(
                "Do you want to first stop these jobs?",
                default = 'n',
                yes = kw.get('yes', False),
                noask = kw.get('noask', False)
            ):
                stop_success_tuple = actions['stop'](
                    action = ['jobs'] + [d.daemon_id for d in _filtered_running_daemons],
                    force = True,
                )
                if not stop_success_tuple[0]:
                    warn(
                        "Failed to stop job" + ("s" if len(_filtered_running_daemons) != 1 else '')
                        + items_str([d.daemon_id for d in _filtered_running_daemons])
                        + ".",
                        stack = False
                    )
                    for d in _filtered_running_daemons:
                        names.remove(d.daemon_id)
                        _filtered_daemons.remove(d)
            else:
                info(
                    "Skipping already running job"
                    + ("s" if len(_filtered_running_daemons) != 1 else '') + ' '
                    + items_str([d.daemon_id for d in _filtered_running_daemons]) + '.'
                )
                for d in _filtered_running_daemons:
                    names.remove(d.daemon_id)
                    _filtered_daemons.remove(d)
                    _skipped_daemons.append(d)

        if not _filtered_daemons:
            return len(_skipped_daemons) > 0, "No jobs to start."

        pprint_jobs(_filtered_daemons, nopretty=kw.get('nopretty', False))
        info(
            f"Starting the job"
            + ("s" if len(names) != 1 else "")
            + " " + items_str(names)
            + "..."
        )

    _successes, _failures = [], []
    for _name in names:
        success_tuple, __name = (
            _run_new_job(_name)
            if new_job
            else _run_existing_job(_name)
        )
        if not kw.get('nopretty', False):
            print_tuple(success_tuple)
        _successes.append(_name) if success_tuple[0] else _failures.append(_name)

    msg = (
        (("Successfully started job" + ("s" if len(_successes) != 1 else '')
            + f" {items_str(_successes)}." + ('\n' if _failures else ''))
            if _successes else '')
        + ("Failed to start job" + ("s" if len(_failures) != 1 else '')
            + f" {items_str(_failures)}." if _failures else '')
    )
    return len(_failures) == 0, msg

def _complete_start_jobs(
        action: Optional[List[str]] = None,
        line: str = '',
        **kw
    ) -> List[str]:
    from meerschaum.utils.daemon import get_daemon_ids
    daemon_ids = get_daemon_ids()
    if not action:
        return daemon_ids
    possibilities = []
    _line_end = line.split(' ')[-1]
    for daemon_id in daemon_ids:
        if daemon_id in action:
            continue
        if _line_end == '':
            possibilities.append(daemon_id)
            continue
        if daemon_id.startswith(action[-1]):
            possibilities.append(daemon_id)
    return possibilities


def _start_gui(
        action: Optional[List[str]] = None,
        mrsm_instance: Optional[str] = None,
        port: Optional[int] = None,
        debug: bool = False,
        **kw
    ) -> SuccessTuple:
    """
    Start the Meerschaum GUI application.
    """
    from meerschaum.utils.daemon import Daemon
    from meerschaum.utils.process import run_process
    from meerschaum.utils.venv import venv_exec
    from meerschaum.utils.packages import (
        run_python_package, attempt_import, venv_contains_package, pip_install
    )
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.networking import find_open_ports, is_port_in_use
    from meerschaum.connectors.parse import parse_instance_keys
    from meerschaum._internal.term.tools import is_webterm_running
    import platform
    webview = attempt_import('webview')
    requests = attempt_import('requests')
    import json
    import time

    success, msg = True, "Success"
    host = '127.0.0.1'
    if port is None:
        port = 8765

    if not is_webterm_running(host, port):
        port = next(find_open_ports(port, 9000))

    api_kw = {
        'action': ['webterm'],
        'no_auth': True,
        'port': port,
        'mrsm_instance': str(parse_instance_keys(mrsm_instance)),
        'debug': debug,
        'host': host,
    }
    api_kw_str = json.dumps(json.dumps(api_kw))
    start_tornado_code = (
        "from meerschaum.actions import actions; "
        "import json; "
        f"actions['start'](**json.loads({api_kw_str}))"
    )
    if debug:
        print(start_tornado_code)
    base_url = 'http://' + api_kw['host'] + ':' + str(api_kw['port'])

    process = venv_exec(
        start_tornado_code, as_proc=True, venv=None, debug=debug, capture_output=(not debug)
    )
    timeout = 10
    start = time.perf_counter()
    starting_up = True
    while starting_up:
        starting_up = (time.perf_counter() - start) < timeout
        time.sleep(0.1)
        try:
            request = requests.get(base_url)
            if request:
                break
        except Exception as e:
            if debug:
                dprint(e)
            continue
    if starting_up is False:
        return False, f"The webterm failed to start within {timeout} seconds."

    try:
        webview.create_window(
            'Meerschaum Shell', 
            f'http://127.0.0.1:{port}',
            height = 650,
            width = 1000
        )
        webview.start(debug=debug)
    except Exception as e:
        import traceback
        traceback.print_exc()
        success, msg = False, str(e)
    except KeyboardInterrupt:
        success, msg = True, "Success"
    finally:
        process.kill()
    return success, msg


def _start_webterm(
        port: Optional[int] = None,
        host: Optional[str] = None,
        force: bool = False,
        nopretty: bool = False,
        sysargs: Optional[List[str]] = None,
        **kw
    ) -> SuccessTuple:
    """
    Start the Meerschaum Web Terminal.
    
    Options:
        - `-p`, `--port`
            The port to which the webterm binds.
            Defaults to 8765, and `--force` will choose the next available port.
    
        - `--host`
            The host interface to which the webterm binds.
            Defaults to '127.0.0.1'.
    """
    from meerschaum._internal.term import get_webterm_app_and_manager, tornado, tornado_ioloop
    from meerschaum._internal.term.tools import is_webterm_running
    from meerschaum.utils.networking import find_open_ports, is_port_in_use
    from meerschaum.utils.packages import attempt_import
    from meerschaum.utils.warnings import info

    if host is None:
        host = '127.0.0.1'
    if port is None:
        port = 8765
    if sysargs is None:
        sysargs = ['start', 'webterm']
    tornado_app, term_manager = get_webterm_app_and_manager()

    if is_webterm_running(host, port):
        if force:
            port = next(find_open_ports(port + 1, 9000))
        else:
            return False, (
                f"The webterm is already running at http://{host}:{port}\n\n"
                + "    Include `-f` to start another server on a new port\n"
                + "    or specify a different port with `-p`."
            )

    if not nopretty:
        info(f"Starting the webterm at http://{host}:{port} ...\n    Press CTRL+C to quit.")
    tornado_app.listen(port, host)
    loop = tornado_ioloop.IOLoop.instance()
    try:
        loop.start()
    except KeyboardInterrupt:
        if not nopretty:
            print()
            info("Shutting down webterm...")
    finally:
        term_manager.shutdown()
        loop.close()

    return True, "Success"


def _start_connectors(
        action: Optional[List[str]] = None,
        connector_keys: Optional[List[str]] = None,
        min_seconds: int = 3,
        debug: bool = False,
        **kw
    ) -> SuccessTuple:
    """
    Start polling connectors to verify a connection can be made.
    """
    from meerschaum.connectors.poll import retry_connect
    from meerschaum.connectors.parse import parse_instance_keys
    from meerschaum.utils.pool import get_pool
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.formatting import pprint
    from meerschaum.utils.misc import items_str
    if action is None:
        action = []
    if connector_keys is None:
        connector_keys = []

    unique_keys = list(set(action + connector_keys))
    valid_conns = []
    for keys in unique_keys:
        try:
            conn = parse_instance_keys(keys)
        except Exception as e:
            warn(f"Invalid connector keys: '{keys}'. Skipping...", stack=False)
            continue
        valid_conns.append(conn)

    if not valid_conns:
        return False, "No valid connector keys were provided."


    connected = {}
    def connect(conn):
        success = retry_connect(
            conn,
            retry_wait = min_seconds,
            enforce_chaining = False,
            enforce_login = False,
            print_on_connect = True,
            debug = debug,
        )
        connected[conn] = success
        return success
    
    pool = get_pool()
    try:
        pool.map(connect, valid_conns)
    except KeyboardInterrupt:
        pass

    ### If a KeyboardInterrupt stopped a connection, mark as `False`.
    for conn in valid_conns:
        if conn not in connected:
            connected[conn] = False

    successes = [conn for conn, success in connected.items() if success]
    fails = [conn for conn, success in connected.items() if not success]

    success = len(fails) == 0
    msg = (
        f"Successfully started connector" + ('s' if len(successes) != 1 else '')
        + ' ' + items_str(successes) + '.'
    ) if success else f"Failed to start {len(fails)} connectors."
    if len(fails) > 0:
        msg += (
            f"\n    Failed to start connector" + ('s' if len(fails) != 1 else '')
            + ' ' + items_str(fails) + '.'
        )

    return success, msg


def _complete_start_connectors(**kw) -> List[str]:
    """
    Return a list of connectors.
    """
    from meerschaum.actions.show import _complete_show_connectors
    return _complete_show_connectors(**kw)


### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.utils.misc import choices_docstring as _choices_docstring
start.__doc__ += _choices_docstring('start')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions useful when chaining API instances together.
"""

from __future__ import annotations

DISALLOW_CHAINING_MESSAGE = """API chaining is not enabled on this Meerschaum instance!

If you manage this server and want to allow this API instance be a parent instance for other API instances,
run `edit config system` and search for `permissions`.
Under 'api:permissions:chaining', change the value of `child_apis` to `true`, and restart the API.

It is highly encouraged that you serve this API over HTTPS! Sensitive data like pipe data and salted
password hashes will be shared between APIs."""

def check_allow_chaining() -> bool:
    """Consult the configuration to see if"""
    from meerschaum.config import get_config
    return get_config('system', 'api', 'permissions', 'chaining', 'child_apis')




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Get the currently registered connectors.
"""

import fastapi
from fastapi import Body, HTTPException
from meerschaum.api import app, endpoints, private, no_auth, manager
from meerschaum.utils.typing import Optional, Dict, List, Union

endpoint = endpoints['connectors']

@app.get(endpoint, tags=['Connectors'])
def get_connectors(
        type : Optional[str] = None,
        curr_user = (
            fastapi.Depends(manager) if not no_auth else None
        ),
    ) -> Union[Dict[str, List[str]], List[str]]:
    """
    Return the keys of the registered connectors.

    Parameters
    ----------
    type:
        If a `type` is specified, return the list of connectors that belong to that type.
        Otherwise, return a dictionary of types that map to lists of labels.
        Defaults to `None`.
    """
    from meerschaum.config import get_config
    if type is not None and type not in get_config('meerschaum', 'connectors'):
        raise HTTPException(status_code=404, detail=f"No connectors of type '{type}'.")
    types = []
    if type is not None:
        types.append(type)
    else:
        for t in get_config('meerschaum', 'connectors'):
            types.append(t)

    response_dict = {}
    for t in types:
        response_dict[t] = list(get_config('meerschaum', 'connectors', t))
        if 'default' in response_dict[t]:
            response_dict[t].remove('default')
    if type is not None:
        return response_dict[type]
    return response_dict

@app.get(endpoint + "/{type}", tags=['Connectors'])
def get_connectors_by_type(
        type : str,
        curr_user = (
            fastapi.Depends(manager) if not no_auth else None
        ),
    ):
    """
    Convenience method for `get_connectors()`.
    """
    return get_connectors(type)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Utility functions for working with DataFrames.
"""

from __future__ import annotations
from meerschaum.utils.typing import (
    Optional, Dict, Any, List, Hashable, Generator,
    Iterator, Iterable,
)


def add_missing_cols_to_df(df: 'pd.DataFrame', dtypes: Dict[str, Any]) -> pd.DataFrame:
    """
    Add columns from the dtypes dictionary as null columns to a new DataFrame.

    Parameters
    ----------
    df: pd.DataFrame
        The dataframe we should copy and add null columns.

    dtypes:
        The data types dictionary which may contain keys not present in `df.columns`.

    Returns
    -------
    A new `DataFrame` with the keys from `dtypes` added as null columns.
    If `df.dtypes` is the same as `dtypes`, then return a reference to `df`.
    NOTE: This will not ensure that dtypes are enforced!

    Examples
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame([{'a': 1}])
    >>> dtypes = {'b': 'Int64'}
    >>> add_missing_cols_to_df(df, dtypes)
          a  b
       0  1  <NA>
    >>> add_missing_cols_to_df(df, dtypes).dtypes
    a    int64
    b    Int64
    dtype: object
    >>> add_missing_cols_to_df(df, {'a': 'object'}).dtypes
    a    int64
    dtype: object
    >>> 
    """
    if set(df.columns) == set(dtypes):
        return df

    import traceback
    from meerschaum.utils.packages import import_pandas, attempt_import
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.dtypes import to_pandas_dtype
    pandas = attempt_import('pandas')
    
    def build_series(dtype: str):
        return pandas.Series([], dtype=to_pandas_dtype(dtype))

    assign_kwargs = {
        str(col): build_series(str(typ))
        for col, typ in dtypes.items()
        if col not in df.columns
    }
    return df.assign(**assign_kwargs)


def filter_unseen_df(
        old_df: 'pd.DataFrame',
        new_df: 'pd.DataFrame',
        dtypes: Optional[Dict[str, Any]] = None,
        debug: bool = False,
    ) -> 'pd.DataFrame':
    """
    Left join two DataFrames to find the newest unseen data.

    Parameters
    ----------
    old_df: 'pd.DataFrame'
        The original (target) dataframe. Acts as a filter on the `new_df`.
        
    new_df: 'pd.DataFrame'
        The fetched (source) dataframe. Rows that are contained in `old_df` are removed.
        
    dtypes: Optional[Dict[str, Any]], default None
        Optionally specify the datatypes of the dataframe.

    debug: bool, default False
        Verbosity toggle.

    Returns
    -------
    A pandas dataframe of the new, unseen rows in `new_df`.

    Examples
    --------
    ```python
    >>> import pandas as pd
    >>> df1 = pd.DataFrame({'a': [1,2]})
    >>> df2 = pd.DataFrame({'a': [2,3]})
    >>> filter_unseen_df(df1, df2)
       a
    0  3

    ```

    """
    if old_df is None:
        return new_df

    import json
    import functools
    import traceback
    from decimal import Decimal
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.packages import import_pandas, attempt_import
    from meerschaum.utils.dtypes import to_pandas_dtype, are_dtypes_equal, attempt_cast_to_numeric
    pd = import_pandas(debug=debug)
    is_dask = 'dask' in new_df.__module__
    if is_dask:
        pandas = attempt_import('pandas')
        dd = attempt_import('dask.dataframe')
        merge = dd.merge
        NA = pandas.NA
    else:
        merge = pd.merge
        NA = pd.NA

    new_df_dtypes = dict(new_df.dtypes)
    old_df_dtypes = dict(old_df.dtypes)

    same_cols = set(new_df.columns) == set(old_df.columns)
    if not same_cols:
        new_df = add_missing_cols_to_df(new_df, old_df_dtypes)
        old_df = add_missing_cols_to_df(old_df, new_df_dtypes)

    ### Edge case: two empty lists cast to DFs.
    elif len(new_df.columns) == 0:
        return new_df

    try:
        ### Order matters when checking equality.
        new_df = new_df[old_df.columns]
    except Exception as e:
        warn(
            "Was not able to cast old columns onto new DataFrame. " +
            f"Are both DataFrames the same shape? Error:\n{e}",
            stacklevel = 3,
        )
        return new_df[list(new_df_dtypes.keys())]

    ### assume the old_df knows what it's doing, even if it's technically wrong.
    if dtypes is None:
        dtypes = {col: str(typ) for col, typ in old_df.dtypes.items()}

    dtypes = {
        col: to_pandas_dtype(typ)
        for col, typ in dtypes.items()
        if col in new_df_dtypes and col in old_df_dtypes
    }
    for col, typ in new_df_dtypes.items():
        if col not in dtypes:
            dtypes[col] = typ
    
    cast_cols = True
    try:
        new_df = new_df.astype(dtypes)
        cast_cols = False
    except Exception as e:
        warn(
            f"Was not able to cast the new DataFrame to the given dtypes.\n{e}"
        )

    new_numeric_cols_existing = get_numeric_cols(new_df)
    old_numeric_cols = get_numeric_cols(old_df)
    for col, typ in {k: v for k, v in dtypes.items()}.items():
        if not are_dtypes_equal(new_df_dtypes.get(col, 'None'), old_df_dtypes.get(col, 'None')):
            new_is_float = are_dtypes_equal(new_df_dtypes.get(col, 'None'), 'float')
            new_is_int = are_dtypes_equal(new_df_dtypes.get(col, 'None'), 'int')
            new_is_numeric = col in new_numeric_cols_existing
            old_is_float = are_dtypes_equal(old_df_dtypes.get(col, 'None'), 'float')
            old_is_int = are_dtypes_equal(old_df_dtypes.get(col, 'None'), 'int')
            old_is_numeric = col in old_numeric_cols

            if (
                (new_is_float or new_is_int or new_is_numeric)
                and
                (old_is_float or old_is_int or old_is_numeric)
            ):
                dtypes[col] = attempt_cast_to_numeric
                cast_cols = True
                continue

            ### Fallback to object if the types don't match.
            warn(
                f"Detected different types for '{col}' "
                + f"({new_df_dtypes.get(col, None)} vs {old_df_dtypes.get(col, None)}), "
                + "falling back to 'object'..."
            )
            dtypes[col] = 'object'
            cast_cols = True

    if cast_cols:
        for col, dtype in dtypes.items():
            if col in new_df.columns:
                try:
                    new_df[col] = (
                        new_df[col].astype(dtype)
                        if not callable(dtype)
                        else new_df[col].apply(dtype)
                    )
                except Exception as e:
                    warn(f"Was not able to cast column '{col}' to dtype '{dtype}'.\n{e}")

    serializer = functools.partial(json.dumps, sort_keys=True, separators=(',', ':'), default=str)
    new_json_cols = get_json_cols(new_df)
    old_json_cols = get_json_cols(old_df)
    json_cols = set(new_json_cols + old_json_cols)
    for json_col in old_json_cols:
        old_df[json_col] = old_df[json_col].apply(serializer)
    for json_col in new_json_cols:
        new_df[json_col] = new_df[json_col].apply(serializer)

    new_numeric_cols = get_numeric_cols(new_df)
    numeric_cols = set(new_numeric_cols + old_numeric_cols)
    for numeric_col in old_numeric_cols:
        old_df[numeric_col] = old_df[numeric_col].apply(
            lambda x: f'{x:f}' if isinstance(x, Decimal) else x
        )
    for numeric_col in new_numeric_cols:
        new_df[numeric_col] = new_df[numeric_col].apply(
            lambda x: f'{x:f}' if isinstance(x, Decimal) else x
        )

    joined_df = merge(
        new_df.fillna(NA),
        old_df.fillna(NA),
        how = 'left',
        on = None,
        indicator = True,
    )
    changed_rows_mask = (joined_df['_merge'] == 'left_only')

    delta_df = joined_df[
        list(new_df_dtypes.keys())
    ][
        changed_rows_mask
    ].reset_index(drop=True)

    for json_col in json_cols:
        if json_col not in delta_df.columns:
            continue
        try:
            delta_df[json_col] = delta_df[json_col].apply(json.loads)
        except Exception as e:
            warn(f"Unable to deserialize JSON column '{json_col}':\n{traceback.format_exc()}")

    for numeric_col in numeric_cols:
        if numeric_col not in delta_df.columns:
            continue
        try:
            delta_df[numeric_col] = delta_df[numeric_col].apply(attempt_cast_to_numeric)
        except Exception as e:
            warn(f"Unable to parse numeric column '{numeric_col}':\n{traceback.format_exc()}")

    return delta_df


def parse_df_datetimes(
        df: 'pd.DataFrame',
        ignore_cols: Optional[Iterable[str]] = None,
        chunksize: Optional[int] = None,
        dtype_backend: str = 'numpy_nullable',
        debug: bool = False,
    ) -> 'pd.DataFrame':
    """
    Parse a pandas DataFrame for datetime columns and cast as datetimes.

    Parameters
    ----------
    df: pd.DataFrame
        The pandas DataFrame to parse.

    ignore_cols: Optional[Iterable[str]], default None
        If provided, do not attempt to coerce these columns as datetimes.

    chunksize: Optional[int], default None
        If the pandas implementation is `'dask'`, use this chunksize for the distributed dataframe.

    dtype_backend: str, default 'numpy_nullable'
        If `df` is not a DataFrame and new one needs to be constructed,
        use this as the datatypes backend.
        Accepted values are 'numpy_nullable' and 'pyarrow'.
        
    debug: bool, default False
        Verbosity toggle.

    Returns
    -------
    A new pandas DataFrame with the determined datetime columns
    (usually ISO strings) cast as datetimes.

    Examples
    --------
    ```python
    >>> import pandas as pd
    >>> df = pd.DataFrame({'a': ['2022-01-01 00:00:00']}) 
    >>> df.dtypes
    a    object
    dtype: object
    >>> df = parse_df_datetimes(df)
    >>> df.dtypes
    a    datetime64[ns]
    dtype: object

    ```

    """
    from meerschaum.utils.packages import import_pandas, attempt_import
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import warn
    import traceback
    pd = import_pandas()
    pandas = attempt_import('pandas')
    pd_name = pd.__name__
    using_dask = 'dask' in pd_name
    df_is_dask = (hasattr(df, '__module__') and 'dask' in df.__module__)
    dask_dataframe = None
    if using_dask or df_is_dask:
        npartitions = chunksize_to_npartitions(chunksize)
        dask_dataframe = attempt_import('dask.dataframe')

    ### if df is a dict, build DataFrame
    if isinstance(df, pandas.DataFrame):
        pdf = df
    elif df_is_dask and isinstance(df, dask_dataframe.DataFrame):
        pdf = get_first_valid_dask_partition(df)
    else:
        if debug:
            dprint(f"df is of type '{type(df)}'. Building {pd.DataFrame}...")

        if using_dask:
            if isinstance(df, list):
                keys = set()
                for doc in df:
                    for key in doc:
                        keys.add(key)
                df = pd.DataFrame.from_dict(
                    {
                        k: [
                            doc.get(k, None)
                            for doc in df
                        ] for k in keys
                    },
                    npartitions = npartitions,
                )
            elif isinstance(df, dict):
                df = pd.DataFrame.from_dict(df, npartitions=npartitions)
            elif 'pandas.core.frame.DataFrame' in str(type(df)):
                df = pd.from_pandas(df, npartitions=npartitions)
            else:
                raise Exception("Can only parse dictionaries or lists of dictionaries with Dask.")
            pandas = attempt_import('pandas')
            pdf = get_first_valid_dask_partition(df)

        else:
            df = pd.DataFrame(df).convert_dtypes(dtype_backend=dtype_backend)
            pdf = df

    ### skip parsing if DataFrame is empty
    if len(pdf) == 0:
        if debug:
            dprint(f"df is empty. Returning original DataFrame without casting datetime columns...")
        return df

    ignore_cols = set(
        (ignore_cols or []) + [
            col
            for col, dtype in pdf.dtypes.items() 
            if 'datetime' in str(dtype)
        ]
    )
    cols_to_inspect = [col for col in pdf.columns if col not in ignore_cols]

    if len(cols_to_inspect) == 0:
        if debug:
            dprint(f"All columns are ignored, skipping datetime detection...")
        return df

    ### apply regex to columns to determine which are ISO datetimes
    iso_dt_regex = r'\d{4}-\d{2}-\d{2}.\d{2}\:\d{2}\:\d+'
    dt_mask = pdf[cols_to_inspect].astype(str).apply(
        lambda s: s.str.match(iso_dt_regex).all()
    )

    ### list of datetime column names
    datetime_cols = [col for col in pdf[cols_to_inspect].loc[:, dt_mask]]
    if not datetime_cols:
        if debug:
            dprint("No columns detected as datetimes, returning...")
        return df

    if debug:
        dprint("Converting columns to datetimes: " + str(datetime_cols))

    try:
        if not using_dask:
            df[datetime_cols] = df[datetime_cols].apply(pd.to_datetime, utc=True)
        else:
            df[datetime_cols] = df[datetime_cols].apply(
                pd.to_datetime,
                utc = True,
                axis = 1,
                meta = {
                    col: 'datetime64[ns]'
                    for col in datetime_cols
                }
            )
    except Exception as e:
        warn(
            f"Unable to apply `pd.to_datetime` to {items_str(datetime_cols)}:\n"
            + f"{traceback.format_exc()}"
        )

    for dt in datetime_cols:
        try:
            df[dt] = df[dt].dt.tz_localize(None)
        except Exception as e:
            warn(f"Unable to convert column '{dt}' to naive datetime:\n{traceback.format_exc()}")

    return df


def get_unhashable_cols(df: 'pd.DataFrame') -> List[str]:
    """
    Get the columns which contain unhashable objects from a Pandas DataFrame.

    Parameters
    ----------
    df: pd.DataFrame
        The DataFrame which may contain unhashable objects.

    Returns
    -------
    A list of columns.
    """
    if len(df) == 0:
        return []

    is_dask = 'dask' in df.__module__
    if is_dask:
        from meerschaum.utils.packages import attempt_import
        pandas = attempt_import('pandas')
        df = pandas.DataFrame(get_first_valid_dask_partition(df))
    return [
        col for col, val in df.iloc[0].items()
        if not isinstance(val, Hashable)
    ]


def get_json_cols(df: 'pd.DataFrame') -> List[str]:
    """
    Get the columns which contain unhashable objects from a Pandas DataFrame.

    Parameters
    ----------
    df: pd.DataFrame
        The DataFrame which may contain unhashable objects.

    Returns
    -------
    A list of columns to be encoded as JSON.
    """
    is_dask = 'dask' in df.__module__
    if is_dask:
        df = get_first_valid_dask_partition(df)
    
    if len(df) == 0:
        return []

    cols_indices = {
        col: df[col].first_valid_index()
        for col in df.columns
    }
    return [
        col
        for col, ix in cols_indices.items()
        if (
            ix is not None
            and
            not isinstance(df.loc[ix][col], Hashable)
        )
    ]


def get_numeric_cols(df: 'pd.DataFrame') -> List[str]:
    """
    Get the columns which contain `decimal.Decimal` objects from a Pandas DataFrame.

    Parameters
    ----------
    df: pd.DataFrame
        The DataFrame which may contain decimal objects.

    Returns
    -------
    A list of columns to treat as numerics.
    """
    from decimal import Decimal
    is_dask = 'dask' in df.__module__
    if is_dask:
        df = get_first_valid_dask_partition(df)
    
    if len(df) == 0:
        return []

    cols_indices = {
        col: df[col].first_valid_index()
        for col in df.columns
    }
    return [
        col
        for col, ix in cols_indices.items()
        if (
            ix is not None
            and
            isinstance(df.loc[ix][col], Decimal)
        )
    ]


def enforce_dtypes(
        df: 'pd.DataFrame',
        dtypes: Dict[str, str],
        debug: bool = False,
    ) -> 'pd.DataFrame':
    """
    Enforce the `dtypes` dictionary on a DataFrame.

    Parameters
    ----------
    df: pd.DataFrame
        The DataFrame on which to enforce dtypes.

    dtypes: Dict[str, str]
        The data types to attempt to enforce on the DataFrame.

    debug: bool, default False
        Verbosity toggle.

    Returns
    -------
    The Pandas DataFrame with the types enforced.
    """
    import json
    import traceback
    from decimal import Decimal
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.formatting import pprint
    from meerschaum.config.static import STATIC_CONFIG
    from meerschaum.utils.packages import import_pandas
    from meerschaum.utils.dtypes import (
        are_dtypes_equal,
        to_pandas_dtype,
        is_dtype_numeric,
        attempt_cast_to_numeric,
    )
    df_dtypes = {c: str(t) for c, t in df.dtypes.items()}
    if len(df_dtypes) == 0:
        if debug:
            dprint("Incoming DataFrame has no columns. Skipping enforcement...")
        return df

    pipe_pandas_dtypes = {
        col: to_pandas_dtype(typ)
        for col, typ in dtypes.items()
    }
    json_cols = [
        col
        for col, typ in dtypes.items()
        if typ == 'json'
    ]
    numeric_cols = [
        col
        for col, typ in dtypes.items()
        if typ == 'numeric'
    ]
    df_numeric_cols = get_numeric_cols(df)
    if debug:
        dprint(f"Desired data types:")
        pprint(dtypes)
        dprint(f"Data types for incoming DataFrame:")
        pprint(df_dtypes)

    if json_cols and len(df) > 0:
        if debug:
            dprint(f"Checking columns for JSON encoding: {json_cols}")
        for col in json_cols:
            if col in df.columns:
                try:
                    df[col] = df[col].apply(
                        (
                            lambda x: (
                                json.loads(x)
                                if isinstance(x, str)
                                else x
                            )
                        )
                    )
                except Exception as e:
                    if debug:
                        dprint(f"Unable to parse column '{col}' as JSON:\n{e}")

    if numeric_cols:
        if debug:
            dprint(f"Checking for numerics: {numeric_cols}")
        for col in numeric_cols:
            if col in df.columns:
                try:
                    df[col] = df[col].apply(attempt_cast_to_numeric)
                except Exception as e:
                    if debug:
                        dprint(f"Unable to parse column '{col}' as NUMERIC:\n{e}")

    if are_dtypes_equal(df_dtypes, pipe_pandas_dtypes):
        if debug:
            dprint(f"Data types match. Exiting enforcement...")
        return df

    common_dtypes = {}
    common_diff_dtypes = {}
    for col, typ in pipe_pandas_dtypes.items():
        if col in df_dtypes:
            common_dtypes[col] = typ
            if not are_dtypes_equal(typ, df_dtypes[col]):
                common_diff_dtypes[col] = df_dtypes[col]

    if debug:
        dprint(f"Common columns with different dtypes:")
        pprint(common_diff_dtypes)

    detected_dt_cols = {}
    for col, typ in common_diff_dtypes.items():
        if 'datetime' in typ and 'datetime' in common_dtypes[col]:
            df_dtypes[col] = typ
            detected_dt_cols[col] = (common_dtypes[col], common_diff_dtypes[col])
    for col in detected_dt_cols:
        del common_diff_dtypes[col]

    if debug:
        dprint(f"Common columns with different dtypes (after dates):")
        pprint(common_diff_dtypes)

    if are_dtypes_equal(df_dtypes, pipe_pandas_dtypes):
        if debug:
            dprint(
                "The incoming DataFrame has mostly the same types, skipping enforcement."
                + f"The only detected difference was in the following datetime columns.\n"
                + "    Timezone information may be stripped."
            )
            pprint(detected_dt_cols)
        return df

    for col, typ in {k: v for k, v in common_diff_dtypes.items()}.items():
        previous_typ = common_dtypes[col]
        mixed_numeric_types = (is_dtype_numeric(typ) and is_dtype_numeric(previous_typ))
        explicitly_float = are_dtypes_equal(dtypes.get(col, 'object'), 'float')
        explicitly_numeric = dtypes.get(col, 'numeric') == 'numeric'
        cast_to_numeric = (
            explicitly_numeric
            or col in df_numeric_cols
            or (mixed_numeric_types and not explicitly_float)
        )
        if cast_to_numeric:
            common_dtypes[col] = attempt_cast_to_numeric
            common_diff_dtypes[col] = attempt_cast_to_numeric

    for d in common_diff_dtypes:
        t = common_dtypes[d]
        if debug:
            dprint(f"Casting column {d} to dtype {t}.")
        try:
            df[d] = (
                df[d].apply(t)
                if callable(t)
                else df[d].astype(t)
            )
        except Exception as e:
            if debug:
                dprint(f"Encountered an error when casting column {d} to type {t}:\n{e}")
            if 'int' in str(t).lower():
                try:
                    df[d] = df[d].astype('float64').astype(t)
                except Exception as e:
                    if debug:
                        dprint(f"Was unable to convert to float then {t}.")
    return df


def get_datetime_bound_from_df(
        df: Union['pd.DataFrame', dict, list],
        datetime_column: str,
        minimum: bool = True,
    ) -> Union[int, 'datetime.datetime', None]:
    """
    Return the minimum or maximum datetime (or integer) from a DataFrame.

    Parameters
    ----------
    df: pd.DataFrame
        The DataFrame, list, or dict which contains the range axis.

    datetime_column: str
        The name of the datetime (or int) column.

    minimum: bool
        Whether to return the minimum (default) or maximum value.

    Returns
    -------
    The minimum or maximum datetime value in the dataframe, or `None`.
    """
    if not datetime_column:
        return None

    def compare(a, b):
        if a is None:
            return b
        if b is None:
            return a
        if minimum:
            return a if a < b else b
        return a if a > b else b

    if isinstance(df, list):
        if len(df) == 0:
            return None
        best_yet = df[0].get(datetime_column, None)
        for doc in df:
            val = doc.get(datetime_column, None)
            best_yet = compare(best_yet, val)
        return best_yet

    if isinstance(df, dict):
        if datetime_column not in df:
            return None
        best_yet = df[datetime_column][0]
        for val in df[datetime_column]:
            best_yet = compare(best_yet, val)
        return best_yet

    if 'DataFrame' in str(type(df)):
        if datetime_column not in df.columns:
            return None
        return (
            df[datetime_column].min(skipna=True)
            if minimum
            else df[datetime_column].max(skipna=True)
        )

    return None


def df_is_chunk_generator(df: Any) -> bool:
    """
    Determine whether to treat `df` as a chunk generator.

    Note this should only be used in a context where generators are expected,
    as it will return `True` for any iterable.

    Parameters
    ----------
    The DataFrame or chunk generator to evaluate.

    Returns
    -------
    A `bool` indicating whether to treat `df` as a generator.
    """
    return (
        not isinstance(df, (dict, list, str))
        and 'DataFrame' not in str(type(df))
        and isinstance(df, (Generator, Iterable, Iterator))
    )


def chunksize_to_npartitions(chunksize: Optional[int]) -> int:
    """
    Return the Dask `npartitions` value for a given `chunksize`.
    """
    if chunksize == -1:
        from meerschaum.config import get_config
        chunksize = get_config('system', 'connectors', 'sql', 'chunksize')
    if chunksize is None:
        return 1
    return -1 * chunksize


def df_from_literal(
        pipe: Optional['meerschaum.Pipe'] = None,
        literal: str = None,
        debug: bool = False
    ) -> 'pd.DataFrame':
    """
    Construct a dataframe from a literal value, using the pipe's datetime and value column names.

    Parameters
    ----------
    pipe: Optional['meerschaum.Pipe'], default None
        The pipe which will consume the literal value.

    Returns
    -------
    A 1-row pandas DataFrame from with the current UTC timestamp as the datetime columns
    and the literal as the value.
    """
    from meerschaum.utils.packages import import_pandas
    from meerschaum.utils.warnings import error, warn
    from meerschaum.utils.debug import dprint

    if pipe is None or literal is None:
        error("Please provide a Pipe and a literal value")
    ### this will raise an error if the columns are undefined
    dt_name, val_name = pipe.get_columns('datetime', 'value')

    val = literal
    if isinstance(literal, str):
        if debug:
            dprint(f"Received literal string: '{literal}'")
        import ast
        try:
            val = ast.literal_eval(literal)
        except Exception as e:
            warn(
                "Failed to parse value from string:\n" + f"{literal}" +
                "\n\nWill cast as a string instead."\
            )
            val = literal

    import datetime
    now = datetime.datetime.utcnow()

    pd = import_pandas()
    return pd.DataFrame({dt_name: [now], val_name: [val]})


def get_first_valid_dask_partition(ddf: 'dask.dataframe.DataFrame') -> Union['pd.DataFrame', None]:
    """
    Return the first valid Dask DataFrame partition (if possible).
    """
    pdf = None
    for partition in ddf.partitions:
        try:
            pdf = partition.compute()
        except Exception as e:
            continue
        if len(pdf) > 0:
            return pdf
    return ddf.compute()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define file paths
"""

from __future__ import annotations

from pathlib import Path
import os, platform, sys, json
from meerschaum.utils.typing import Union
from meerschaum.config.static import STATIC_CONFIG

DOT_CONFIG_DIR_PATH = Path(
    os.environ.get('XDG_CONFIG_HOME', Path.home() / '.config')
    if platform.system() != 'Windows'
    else os.environ.get('AppData', Path.home() / 'AppData' / 'Roaming')
)

DEFAULT_ROOT_DIR_PATH = (
    (DOT_CONFIG_DIR_PATH / 'meerschaum')
    if platform.system() != 'Windows'
    else (DOT_CONFIG_DIR_PATH / 'Meerschaum')
)


ENVIRONMENT_ROOT_DIR = STATIC_CONFIG['environment']['root']
if ENVIRONMENT_ROOT_DIR in os.environ:
    _ROOT_DIR_PATH = Path(os.environ[ENVIRONMENT_ROOT_DIR]).resolve()
    if not _ROOT_DIR_PATH.exists():
        print(
            f"Invalid root directory '{str(_ROOT_DIR_PATH)}' set for " +
            f"environment variable '{ENVIRONMENT_ROOT_DIR}'.\n" +
            f"Please enter a valid path for {ENVIRONMENT_ROOT_DIR}.",
            file = sys.stderr,
        )
        sys.exit(1)
else:
    _ROOT_DIR_PATH = DEFAULT_ROOT_DIR_PATH

ENVIRONMENT_PLUGINS_DIR = STATIC_CONFIG['environment']['plugins']
if ENVIRONMENT_PLUGINS_DIR in os.environ:
    try:
        PLUGINS_DIR_PATHS = (
            [
                Path(path).resolve()
                for path in json.loads(os.environ[ENVIRONMENT_PLUGINS_DIR])
            ] if os.environ[ENVIRONMENT_PLUGINS_DIR].lstrip().startswith('[')
            else [
                Path(path_str).resolve()
                for path_str in os.environ[ENVIRONMENT_PLUGINS_DIR].split(':')
                if path_str
            ]
        )
    except Exception as e:
        PLUGINS_DIR_PATHS = []

    if not PLUGINS_DIR_PATHS:
        print(
            "Invalid plugins directories set for "
            f"environment variable '{ENVIRONMENT_PLUGINS_DIR}'.\n\n"
            f"Set this to a colon-separated path string:\n\n"
            f"`export {ENVIRONMENT_PLUGINS_DIR}=./plugins:/another/path/to/plugins`\n\n"
            "or a JSON-encoded path list:\n\n"
            f"`export {ENVIRONMENT_PLUGINS_DIR}=" + "'[\"./plugins\", \"/another/path/to/plugins\"]'`"
            f"",
        )
        sys.exit(1)
else:
    PLUGINS_DIR_PATHS = [_ROOT_DIR_PATH / 'plugins']

### Remove duplicate plugins paths.
_seen_plugins_paths, _plugins_paths_to_remove = set(), set()
for _plugin_path in PLUGINS_DIR_PATHS:
    if _plugin_path in _seen_plugins_paths:
        _plugins_paths_to_remove.add(_plugin_path)
    _seen_plugins_paths.add(_plugin_path)
for _plugin_path in _plugins_paths_to_remove:
    PLUGINS_DIR_PATHS.remove(_plugin_path)

ENVIRONMENT_VENVS_DIR = STATIC_CONFIG['environment']['venvs']
if ENVIRONMENT_VENVS_DIR in os.environ:
    VENVS_DIR_PATH = Path(os.environ[ENVIRONMENT_VENVS_DIR]).resolve()
    if not VENVS_DIR_PATH.exists():
        try:
            VENVS_DIR_PATH.mkdir(parents=True, exist_ok=True)
        except Exception as e:
            print(
                f"Invalid path set for environment variable '{ENVIRONMENT_VENVS_DIR}':\n"
                + f"{VENVS_DIR_PATH}"
            )
            VENVS_DIR_PATH = (_ROOT_DIR_PATH / 'venvs').resolve()
            print(f"Will use the following path for venvs instead:\n{VENVS_DIR_PATH}")
else:
    VENVS_DIR_PATH = _ROOT_DIR_PATH / 'venvs'

paths = {
    'PACKAGE_ROOT_PATH'              : str(Path(__file__).parent.parent.resolve()),
    'ROOT_DIR_PATH'                  : str(_ROOT_DIR_PATH),
    'VIRTENV_RESOURCES_PATH'         : str(VENVS_DIR_PATH),
    'CONFIG_DIR_PATH'                : ('{ROOT_DIR_PATH}', 'config'),
    'DEFAULT_CONFIG_DIR_PATH'        : ('{ROOT_DIR_PATH}', 'default_config'),
    'PATCH_DIR_PATH'                 : ('{ROOT_DIR_PATH}', 'patch_config'),
    'PERMANENT_PATCH_DIR_PATH'       : ('{ROOT_DIR_PATH}', 'permanent_patch_config'),
    'INTERNAL_RESOURCES_PATH'        : ('{ROOT_DIR_PATH}', '.internal'),

    'STACK_RESOURCES_PATH'           : ('{ROOT_DIR_PATH}', 'stack'),
    'STACK_COMPOSE_FILENAME'         : 'docker-compose.yaml',
    'STACK_COMPOSE_PATH'             : ('{STACK_RESOURCES_PATH}', '{STACK_COMPOSE_FILENAME}'),
    'STACK_ENV_FILENAME'             : '.env',
    'STACK_ENV_PATH'                 : ('{STACK_RESOURCES_PATH}', '{STACK_ENV_FILENAME}'),

    'SHELL_RESOURCES_PATH'           : ('{ROOT_DIR_PATH}', ),
    'SHELL_HISTORY_PATH'             : ('{SHELL_RESOURCES_PATH}', '.mrsm_history'),

    'API_RESOURCES_PATH'             : ('{PACKAGE_ROOT_PATH}', 'api', 'resources'),
    'API_STATIC_PATH'                : ('{API_RESOURCES_PATH}', 'static'),
    'API_TEMPLATES_PATH'             : ('{API_RESOURCES_PATH}', 'templates'),
    'API_CONFIG_RESOURCES_PATH'      : ('{ROOT_DIR_PATH}', 'api'),
    'API_SECRET_KEY_PATH'            : ('{API_CONFIG_RESOURCES_PATH}', '.api_secret_key'),
    'API_UVICORN_RESOURCES_PATH'     : ('{API_CONFIG_RESOURCES_PATH}', 'uvicorn'),
    'API_UVICORN_CONFIG_PATH'        : ('{API_UVICORN_RESOURCES_PATH}', '.thread_config.json'),

    'CACHE_RESOURCES_PATH'           : ('{ROOT_DIR_PATH}', '.cache'),
    'PIPES_CACHE_RESOURCES_PATH'     : ('{CACHE_RESOURCES_PATH}', 'pipes'),
    'USERS_CACHE_RESOURCES_PATH'     : ('{CACHE_RESOURCES_PATH}', 'users'),

    'PLUGINS_RESOURCES_PATH'         : ('{INTERNAL_RESOURCES_PATH}', 'plugins'),
    'PLUGINS_INTERNAL_LOCK_PATH'     : ('{INTERNAL_RESOURCES_PATH}', 'plugins.lock'),
    'PLUGINS_ARCHIVES_RESOURCES_PATH': ('{PLUGINS_RESOURCES_PATH}', '.archives'),
    'PLUGINS_TEMP_RESOURCES_PATH'    : ('{PLUGINS_RESOURCES_PATH}', '.tmp'),
    'PLUGINS_INIT_PATH'              : ('{PLUGINS_RESOURCES_PATH}', '__init__.py'),

    'SQLITE_RESOURCES_PATH'          : ('{ROOT_DIR_PATH}', 'sqlite'),
    'SQLITE_DB_PATH'                 : ('{SQLITE_RESOURCES_PATH}', 'mrsm_local.db'),

    'DUCKDB_RESOURCES_PATH'          : ('{ROOT_DIR_PATH}', 'duckdb'),
    'DUCKDB_PATH'                    : ('{DUCKDB_RESOURCES_PATH}', 'duck.db'),

    'GRAFANA_RESOURCES_PATH'         : ('{STACK_RESOURCES_PATH}', 'grafana', 'resources'),
    'GRAFANA_DATASOURCE_PATH'        : (
        '{GRAFANA_RESOURCES_PATH}', 'provisioning', 'datasources', 'datasource.yaml'
    ),
    'GRAFANA_DASHBOARD_PATH'         : (
        '{GRAFANA_RESOURCES_PATH}', 'provisioning', 'dashboards', 'dashboard.yaml'
    ),
    'MOSQUITTO_RESOURCES_PATH'       : ('{STACK_RESOURCES_PATH}', 'mosquitto', 'resources'),
    'MOSQUITTO_CONFIG_PATH'          : ('{MOSQUITTO_RESOURCES_PATH}', 'mosquitto.conf'),

    'PORTABLE_CHECK_READLINE_PATH'   : ('{SHELL_RESOURCES_PATH}', '.readline_attempted_install'),

    'DAEMON_RESOURCES_PATH'          : ('{ROOT_DIR_PATH}', 'jobs'),
    'LOGS_RESOURCES_PATH'            : ('{ROOT_DIR_PATH}', 'logs'),
}

def set_root(root: Union[Path, str]):
    """Modify the value of `ROOT_DIR_PATH`."""
    paths['ROOT_DIR_PATH'] = Path(root).resolve()
    for path_name, path_parts in paths.items():
        if isinstance(path_parts, tuple) and path_parts[0] == '{ROOT_DIR_PATH}':
            globals()[path_name] = __getattr__(path_name)

def __getattr__(name: str) -> Path:
    if name not in paths:
        if name not in globals():
            raise AttributeError(f"Could not import '{name}'.")
        return globals()[name]

    if isinstance(paths[name], (list, tuple)) and len(paths[name]) > 0:
        ### recurse through paths to create resource directories.
        parts = []
        for p in paths[name]:
            if str(p).startswith('{') and str(p).endswith('}'):
                parts.append(__getattr__(p[1:-1]))
            else:
                parts.append(p)
        path = Path(os.path.join(*parts))
    else:
        path = Path(paths[name])

    ### Create directories or touch files.
    if name.endswith('RESOURCES_PATH') or name == 'CONFIG_DIR_PATH':
        path.mkdir(parents=True, exist_ok=True)
    elif 'FILENAME' in name:
        path = str(path)

    return path




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Manage plugins via the API connector
"""

from __future__ import annotations
from meerschaum.utils.typing import Union, Any, Optional, SuccessTuple, Mapping, Sequence

def plugin_r_url(
        plugin: Union[meerschaum.core.Plugin, str]
    ) -> str:
    """Generate a relative URL path from a Plugin."""
    from meerschaum.config.static import STATIC_CONFIG
    return f"{STATIC_CONFIG['api']['endpoints']['plugins']}/{plugin}"

def register_plugin(
        self,
        plugin: meerschaum.core.Plugin,
        make_archive: bool = True,
        debug: bool = False,
    ) -> SuccessTuple:
    """Register a plugin and upload its archive."""
    import json
    archive_path = plugin.make_tar(debug=debug) if make_archive else plugin.archive_path
    file_pointer = open(archive_path, 'rb')
    files = {'archive': file_pointer}
    metadata = {
        'version': plugin.version,
        'attributes': json.dumps(plugin.attributes),
    }
    r_url = plugin_r_url(plugin)
    try:
        response = self.post(r_url, files=files, params=metadata, debug=debug)
    except Exception as e:
        return False, f"Failed to register plugin '{plugin}'."
    finally:
        file_pointer.close()

    try:
        success, msg = json.loads(response.text)
    except Exception as e:
        return False, response.text

    return success, msg

def install_plugin(
        self,
        name: str,
        force: bool = False,
        debug: bool = False
    ) -> SuccessTuple:
    """Download and attempt to install a plugin from the API."""
    import os, pathlib, json
    from meerschaum.core import Plugin
    from meerschaum.config._paths import PLUGINS_TEMP_RESOURCES_PATH
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.packages import attempt_import
    binaryornot_check = attempt_import('binaryornot.check', lazy=False)
    r_url = plugin_r_url(name)
    dest = pathlib.Path(os.path.join(PLUGINS_TEMP_RESOURCES_PATH, name + '.tar.gz'))
    if debug:
        dprint(f"Fetching from '{self.url + r_url}' to '{dest}'...")
    archive_path = self.wget(r_url, dest, debug=debug) 
    is_binary = binaryornot_check.is_binary(str(archive_path))
    if not is_binary:
        fail_msg = f"Failed to download binary for plugin '{name}'."
        try:
            with open(archive_path, 'r') as f:
                j = json.load(f)
            if isinstance(j, list):
                success, msg = tuple(j)
            elif isinstance(j, dict) and 'detail' in j:
                success, msg = False, fail_msg
        except Exception as e:
            success, msg = False, fail_msg
        return success, msg
    plugin = Plugin(name, archive_path=archive_path, repo_connector=self)
    return plugin.install(force=force, debug=debug)

def get_plugins(
        self,
        user_id : Optional[int] = None,
        search_term : Optional[str] = None,
        debug : bool = False
    ) -> Sequence[str]:
    """Return a list of registered plugin names.

    Parameters
    ----------
    user_id :
        If specified, return all plugins from a certain user.
    user_id : Optional[int] :
         (Default value = None)
    search_term : Optional[str] :
         (Default value = None)
    debug : bool :
         (Default value = False)

    Returns
    -------

    """
    import json
    from meerschaum.utils.warnings import warn, error
    from meerschaum.config.static import STATIC_CONFIG
    response = self.get(
        STATIC_CONFIG['api']['endpoints']['plugins'],
        params = {'user_id' : user_id, 'search_term' : search_term},
        use_token = True,
        debug = debug
    )
    if not response:
        return []
    plugins = json.loads(response.text)
    if not isinstance(plugins, list):
        error(response.text)
    return plugins

def get_plugin_attributes(
        self,
        plugin: meerschaum.core.Plugin,
        debug: bool = False
    ) -> Mapping[str, Any]:
    """
    Return a plugin's attributes.
    """
    import json
    from meerschaum.utils.warnings import warn, error
    r_url = plugin_r_url(plugin) + '/attributes'
    response = self.get(r_url, use_token=True, debug=debug)
    attributes = response.json()
    if isinstance(attributes, str) and attributes and attributes[0] == '{':
        try:
            attributes = json.loads(attributes)
        except Exception as e:
            pass
    if not isinstance(attributes, dict):
        error(response.text)
    elif not response and 'detail' in attributes:
        warn(attributes['detail'])
        return {}
    return attributes

def delete_plugin(
        self,
        plugin: meerschaum.core.Plugin,
        debug: bool = False
    ) -> SuccessTuple:
    """Delete a plugin from an API repository."""
    import json
    r_url = plugin_r_url(plugin)
    try:
        response = self.delete(r_url, debug=debug)
    except Exception as e:
        return False, f"Failed to delete plugin '{plugin}'."

    try:
        success, msg = json.loads(response.text)
    except Exception as e:
        return False, response.text

    return success, msg




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Docker Compose stack configuration goes here
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, List, Any, SuccessTuple, Dict

import os
import json
from meerschaum.config._paths import (
    GRAFANA_DATASOURCE_PATH,
    GRAFANA_DASHBOARD_PATH,
    ROOT_DIR_PATH,
)
from meerschaum.config._paths import STACK_COMPOSE_FILENAME, STACK_ENV_FILENAME
from meerschaum.config._paths import CONFIG_DIR_PATH, STACK_ENV_PATH, STACK_COMPOSE_PATH
from meerschaum.config._paths import GRAFANA_DATASOURCE_PATH, GRAFANA_DASHBOARD_PATH

db_port = "MRSM{meerschaum:connectors:sql:main:port}"
db_user = "MRSM{meerschaum:connectors:sql:main:username}"
db_pass = "MRSM{meerschaum:connectors:sql:main:password}"
db_base = "MRSM{meerschaum:connectors:sql:main:database}"

### default localhost, db for docker network
db_hostname = "db"
db_host = 'MRSM{stack:' + str(STACK_COMPOSE_FILENAME) + ':services:db:hostname}'
api_port = "MRSM{meerschaum:connectors:api:main:port}"
api_host = "api"

env_dict = {
    'COMPOSE_PROJECT_NAME' : 'mrsm',
    'TIMESCALEDB_VERSION' : 'latest-pg15-oss',
    'POSTGRES_USER' : f'{db_user}',
    'POSTGRES_PASSWORD' : f'{db_pass}',
    'POSTGRES_DB' : f'{db_base}',
    'MEERSCHAUM_API_HOSTNAME' : f'{api_host}',
    'ALLOW_IP_RANGE' : '0.0.0.0/0',
    'MEERSCHAUM_API_CONFIG_RESOURCES' : '/meerschaum',
}
### apply patch to host config to change hostname to the Docker service name
env_dict['MEERSCHAUM_API_CONFIG'] = json.dumps(
    {
        'meerschaum' : 'MRSM{!meerschaum}',
        'system' : 'MRSM{!system}',
    },
    indent = 4,
).replace(
    '"MRSM{!system}"', 'MRSM{!system}'
).replace(
    '"MRSM{!meerschaum}"', 'MRSM{!meerschaum}',
)

volumes = {
    'api_root': '/meerschaum',
    'meerschaum_db_data': '/var/lib/postgresql/data',
    'grafana_storage': '/var/lib/grafana',
}
networks = {
    'frontend': None,
    'backend': None,
}
env_dict['MEERSCHAUM_API_PATCH'] = json.dumps(
    {
        'meerschaum': {
            'connectors': {
                'sql': {
                    'main': {
                        'host': db_host,
                        'port': 5432,
                    },
                    'local': {
                        'database': volumes['api_root'] + '/sqlite/mrsm_local.db'
                    },
                },
            },
        },
    },
    indent = 4,
)

compose_header = """
##############################################################
#                                                            #
#                   DO NOT EDIT THIS FILE!                   #
#                                                            #
#          Any changes you make will be overwritten.         #
#                                                            #
# Instead, you can change this file's configuration with     #
# `edit config stack` under the docker-compose.yaml section. #
#                                                            #
##############################################################
"""


default_docker_compose_config = {
    'version': '3.9',
    'services': {
        'db': {
            'environment': {
                'TIMESCALEDB_TELEMETRY': 'off',
                'POSTGRES_USER': '<DOLLAR>POSTGRES_USER',
                'POSTGRES_DB': '<DOLLAR>POSTGRES_DB',
                'POSTGRES_PASSWORD': '<DOLLAR>POSTGRES_PASSWORD',
                'ALLOW_IP_RANGE': env_dict['ALLOW_IP_RANGE'],
            },
            'command': 'postgres -c max_connections=1000 -c shared_buffers=1024MB',
            'healthcheck': {
                'test': [
                    'CMD-SHELL', 'pg_isready -d <DOLLAR>POSTGRES_DB -U <DOLLAR>POSTGRES_USER',
                ],
                'interval': '5s',
                'timeout': '3s',
                'retries': 3
            },
            'restart': 'always',
            'image' : 'timescale/timescaledb:' + env_dict['TIMESCALEDB_VERSION'],
            'ports' : [
                f'{db_port}:5432',
            ],
            'hostname' : f'{db_hostname}',
            'volumes' : [
                'meerschaum_db_data' + ':' + volumes['meerschaum_db_data'],
            ],
            'shm_size': '1024m',
            'networks' : [
                'backend',
            ],
        },
        'api': {
            'image': 'bmeares/meerschaum:api',
            'ports': [f'{api_port}:{api_port}'],
            'hostname': f'{api_host}',
            'networks': [
                'frontend',
                'backend',
            ],
            'command': f'start api --production --port {api_port}',
            'healthcheck': {
                'test': [
                    'CMD', 'curl', '--fail', f'http://localhost:{api_port}/healthcheck',
                ],
                'interval': '5s',
                'timeout': '3s',
                'retries': 3
            },
            'environment': {
                'MRSM_CONFIG': env_dict['MEERSCHAUM_API_CONFIG'],
                'MRSM_PATCH': env_dict['MEERSCHAUM_API_PATCH'],
            },
            'restart': 'always',
            'init': True,
            'depends_on': {
                'db': {
                    'condition': 'service_healthy',
                },
            },
            'volumes' : [
                'api_root:' + volumes['api_root'],
            ],
        },
        'grafana': {
            'image': 'grafana/grafana:latest',
            'ports': [
                '3000:3000',
            ],
            'networks': [
                'frontend',
                'backend',
            ],
            'restart': 'always',
            'depends_on': {
                'db': {
                    'condition': 'service_healthy',
                },
            },
            'volumes': [
                'grafana_storage' + ':' + volumes['grafana_storage'],
                ### NOTE: Mount with the 'z' option for SELinux.
                f'{GRAFANA_DATASOURCE_PATH.parent}:/etc/grafana/provisioning/datasources:z,ro',
                f'{GRAFANA_DASHBOARD_PATH.parent}:/etc/grafana/provisioning/dashboards:z,ro',
            ],
            'environment': {
                'GF_SECURITY_ALLOW_EMBEDDING': 'true',
                'GF_ANALYTICS_REPORTING_ENABLED': 'false',
                'GF_AUTH_ANONYMOUS_ENABLED': 'true',
                'GF_AUTH_ANONYMOUS_ORGANIZATION': 'public',
            },
        },
    },
}
default_docker_compose_config['networks'] = networks
default_docker_compose_config['volumes'] = {}
for key in volumes:
    default_docker_compose_config['volumes'][key] = None

default_stack_config = {}
### compose project name (prepends to all services)
default_stack_config['project_name'] = 'mrsm'
compose_filename = os.path.split(STACK_COMPOSE_PATH)[1]
default_stack_config[compose_filename] = default_docker_compose_config
from meerschaum.config.stack.grafana import default_grafana_config
default_stack_config['grafana'] = default_grafana_config
default_stack_config['filetype'] = 'yaml'

### check if configs are in sync

def _sync_stack_files():
    from meerschaum.config._sync import sync_yaml_configs
    sync_yaml_configs(
        CONFIG_DIR_PATH / 'stack.yaml',
        ['stack', STACK_COMPOSE_FILENAME],
        STACK_COMPOSE_PATH,
        substitute = True,
    )
    sync_yaml_configs(
        CONFIG_DIR_PATH / 'stack.yaml',
        ['stack', 'grafana', 'datasource'],
        GRAFANA_DATASOURCE_PATH,
        substitute = True,
    )
    sync_yaml_configs(
        CONFIG_DIR_PATH / 'stack.yaml',
        ['stack', 'grafana', 'dashboard'],
        GRAFANA_DASHBOARD_PATH,
        substitute = True,
    )

NECESSARY_FILES = [STACK_COMPOSE_PATH, GRAFANA_DATASOURCE_PATH, GRAFANA_DASHBOARD_PATH]
def get_necessary_files():
    from meerschaum.config import get_config
    return {
        STACK_COMPOSE_PATH : (
            get_config('stack', STACK_COMPOSE_FILENAME, substitute=True), compose_header
        ),
        GRAFANA_DATASOURCE_PATH : get_config('stack', 'grafana', 'datasource', substitute=True),
        GRAFANA_DASHBOARD_PATH : get_config('stack', 'grafana', 'dashboard', substitute=True),
    }


def write_stack(
        debug: bool = False 
    ):
    """Write Docker Compose configuration files."""
    from meerschaum.config._edit import general_write_yaml_config
    from meerschaum.config._sync import sync_files
    general_write_yaml_config(get_necessary_files(), debug=debug)
    return sync_files(['stack'])
   
def edit_stack(
        action : Optional[List[str]] = None,
        debug : bool = False,
        **kw
    ):
    """Open docker-compose.yaml or .env for editing."""
    from meerschaum.config._edit import general_edit_config
    if action is None:
        action = []
    files = {
        'compose' : STACK_COMPOSE_PATH,
        'docker-compose' : STACK_COMPOSE_PATH,
        'docker-compose.yaml' : STACK_COMPOSE_PATH,
    }
    return general_edit_config(action=action, files=files, default='compose', debug=debug)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Login page for the Web interface.
"""

from __future__ import annotations
import json
from meerschaum.api import permissions_config, endpoints, CHECK_UPDATE
from meerschaum.utils.packages import import_html, import_dcc
html, dcc = import_html(check_update=CHECK_UPDATE), import_dcc(check_update=CHECK_UPDATE)
import dash_bootstrap_components as dbc
from meerschaum.config import get_config

dash_endpoint = endpoints['dash']
allow_user_registration = permissions_config['registration']['users']

registration_div = html.Div(
    id = 'registration-div',
    style = {'height' : '100%'},
    children = (
        [
            dcc.Link(
                'No account? Create one here.',
                href = (dash_endpoint + '/register'),
                refresh = False
            ),
        ] if allow_user_registration
        else [
            dcc.Markdown("""
                #### **Web registration is disabled for security.**
                You can still register users on this instance with a SQL connector.
            """),
            dbc.Button(
                'More information.',
                id = 'show-user-registration-disabled-button',
                color = 'link',
                size = 'sm',
            ),
            dbc.Collapse(
                id = 'user-registration-disabled-collapse',
                children = [
                    dcc.Markdown(
                        "For example, to register user `newuser` on instance `sql:main`:"
                    ),
                    html.Pre(
                        html.Code('mrsm register user newuser -i sql:main', className='codeblock'),
                    ),
                    dcc.Markdown("""
                        To enable online registration, open the `system` configuration file and""" +
                    """ set the permissions to `true`:"""
                    ),
                    html.Pre(
                        html.Code('mrsm edit config system', className='codeblock'),
                    ),
                    html.Br(),
                    dcc.Markdown('The settings file should look something like this:'),
                    html.Pre(
                        html.Code(
                            json.dumps({
                                'api' : {
                                    'permissions' : {
                                        'registration' : {
                                            'users' : True,
                                        },
                                    }
                                }
                            }, indent=2),
                        className='codeblock'),
                    ),
                ]
            ),
        ]
    )
)

layout = dbc.Container([
    html.Br(),
    dbc.Container([
        dcc.Location(id='location-login', refresh=True),
        html.Div([
            dbc.Container(
                html.Img(
                    src = endpoints['dash'] + "/assets/banner_1920x320.png",
                    width = '100%',
                    className = 'center'
                ),
            ),
            html.Br(),
            dbc.Container(id='login-container', children=[
                dbc.Row([
                    dbc.Col(
                    dcc.Input(
                        placeholder = 'Username',
                        type = 'text',
                        id = 'username-input',
                        className = 'form-control',
                        n_submit = 0,
                    )),
                    dbc.Col(dcc.Input(
                        placeholder = 'Password',
                        type = 'password',
                        id = 'password-input',
                        className = 'form-control',
                        n_submit = 0,
                    )),
                ]),
                html.Br(),
                dbc.Row([
                    dbc.Col([
                        html.Button(
                            children = 'Login',
                            n_clicks = 0,
                            type = 'submit',
                            id = 'login-button',
                            className = 'btn btn-primary btn-lg'
                        ),
                    ]),
                ]),
                html.Br(),
                dbc.Row([
                    dbc.Col([
                        registration_div,
                    ]),
                ]),
            ], className='form-group'),
        ]),
    ], className='jumbotron')
])




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Launch into a CLI environment to interact with the SQL Connector
"""

from __future__ import annotations
import os
import json
import copy
### NOTE: This import adds `Iterable` to collections, which is needed by some CLIs.
from meerschaum.utils.typing import SuccessTuple

flavor_clis = {
    'postgresql'  : 'pgcli',
    'timescaledb' : 'pgcli',
    'cockroachdb' : 'pgcli',
    'citus'       : 'pgcli',
    'mysql'       : 'mycli',
    'mariadb'     : 'mycli',
    'percona'     : 'mycli',
    'sqlite'      : 'litecli',
    'mssql'       : 'mssqlcli',
    'duckdb'      : 'gadwall',
}
cli_deps = {
    'pgcli': ['pgspecial', 'pendulum'],
    'mycli': ['cryptography'],
}


def cli(
        self,
        debug: bool = False,
    ) -> SuccessTuple:
    """
    Launch a subprocess for an interactive CLI.
    """
    from meerschaum.utils.venv import venv_exec
    env = copy.deepcopy(dict(os.environ))
    env[f'MRSM_SQL_{self.label.upper()}'] = json.dumps(self.meta)
    cli_code = (
        "import meerschaum as mrsm\n"
        f"conn = mrsm.get_connector('sql:{self.label}')\n"
        f"conn._cli_exit()"
    )
    try:
        _ = venv_exec(cli_code, venv=None, debug=debug, capture_output=False)
    except Exception as e:
        return False, f"[{self}] Failed to start CLI:\n{e}"
    return True, "Success"


def _cli_exit(
        self,
        debug: bool = False
    ) -> SuccessTuple:
    """Launch an interactive CLI for the SQLConnector's flavor."""
    from meerschaum.utils.packages import venv_exec, attempt_import
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import error
    import sys, subprocess, os

    if self.flavor not in flavor_clis:
        return False, f"No CLI available for flavor '{self.flavor}'."

    if self.flavor == 'duckdb':
        gadwall = attempt_import('gadwall', debug = debug, lazy=False)
        gadwall_shell = gadwall.Gadwall(self.database)
        try:
            gadwall_shell.cmdloop()
        except KeyboardInterrupt:
            pass
        return True, "Success"
    elif self.flavor == 'mssql':
        if 'DOTNET_SYSTEM_GLOBALIZATION_INVARIANT' not in os.environ:
            os.environ['DOTNET_SYSTEM_GLOBALIZATION_INVARIANT'] = '1'

    cli_name = flavor_clis[self.flavor]

    ### Install the CLI package and any dependencies.
    cli, cli_main = attempt_import(cli_name, (cli_name + '.main'), lazy=False, debug=debug)
    if cli_name in cli_deps:
        for dep in cli_deps[cli_name]:
            locals()[dep] = attempt_import(dep, lazy=False, warn=False, debug=debug)

    ### NOTE: The `DATABASE_URL` property must be initialized first in case the database is not
    ### yet defined (e.g. 'sql:local').
    cli_arg_str = self.DATABASE_URL
    if self.flavor in ('sqlite', 'duckdb'):
        cli_arg_str = str(self.database)

    ### Define the script to execute to launch the CLI.
    ### The `mssqlcli` script is manually written to avoid telemetry
    ### and because `main.cli()` is not defined.
    launch_cli = f"cli_main.cli(['{cli_arg_str}'])"
    if self.flavor == 'mssql':
        launch_cli = (
            "mssqlclioptionsparser, mssql_cli = attempt_import("
            + "'mssqlcli.mssqlclioptionsparser', 'mssqlcli.mssql_cli', lazy=False)\n"
            + "ms_parser = mssqlclioptionsparser.create_parser()\n"
            + f"ms_options = ms_parser.parse_args(['--server', 'tcp:{self.host},{self.port}', "
            + f"'--database', '{self.database}', "
            + f"'--username', '{self.username}', '--password', '{self.password}'])\n"
            + "ms_object = mssql_cli.MssqlCli(ms_options)\n"
            + "try:\n"
            + "    ms_object.connect_to_database()\n"
            + "    ms_object.run()\n"
            + "finally:\n"
            + "    ms_object.shutdown()"
        )

    try:
        if debug:
            dprint(f'Launching CLI:\n{launch_cli}')
        exec(launch_cli)
        success, msg = True, 'Success'
    except Exception as e:
        success, msg = False, str(e)

    return success, msg



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Insert non-user-editable configuration files here.
"""

import os
import uuid
from typing import Dict, Any
from meerschaum.utils.misc import generate_password

__all__ = ['STATIC_CONFIG']

SERVER_ID: str = os.environ.get('MRSM_SERVER_ID', generate_password(6))
STATIC_CONFIG: Dict[str, Any] = {
    'api': {
        'endpoints': {
            'index': '/',
            'favicon': '/favicon.ico',
            'plugins': '/plugins',
            'pipes': '/pipes',
            'metadata': '/metadata',
            'actions': '/actions',
            'users': '/users',
            'login': '/login',
            'connectors': '/connectors',
            'version': '/version',
            'chaining': '/chaining',
            'websocket': '/ws',
            'dash': '/dash',
            'webterm': '/webterm',
            'webterm_websocket': '/websocket',
            'info': '/info',
            'healthcheck': '/healthcheck',
        },
        'oauth': {
            'token_expires_minutes': 720,
        },
        'webterm_job_name': '_webterm',
    },
    'sql': {
        'internal_schema': '_mrsm_internal',
        'instance_schema': 'mrsm',
    },
    'environment': {
        'config': 'MRSM_CONFIG',
        'patch': 'MRSM_PATCH',
        'root': 'MRSM_ROOT_DIR',
        'plugins': 'MRSM_PLUGINS_DIR',
        'venvs': 'MRSM_VENVS_DIR',
        'runtime': 'MRSM_RUNTIME',
        'work_dir': 'MRSM_WORK_DIR',
        'user': 'MRSM_USER',
        'dep_group': 'MRSM_DEP_GROUP',
        'home': 'MRSM_HOME',
        'src': 'MRSM_SRC',
        'noask': 'MRSM_NOASK',
        'id': 'MRSM_SERVER_ID',
        'uri_regex': r'MRSM_([a-zA-Z0-9]*)_(\d*[a-zA-Z][a-zA-Z0-9-_+]*$)',
        'prefix': 'MRSM_',
    },
    'config': {
        'default_filetype': 'json',
        'symlinks_key': '_symlinks',
    },
    'system': {
        'arguments': {
            'sub_decorators': (
                '[',
                ']'
            ),
            'underscore_standin': '<UNDERSCORE>', ### Temporary replacement for parsing.
            'failure_key': '_argparse_exception',
        },
        'urls': {
            'get-pip.py': 'https://bootstrap.pypa.io/get-pip.py',
        },
        'success': {
            'ignore': (
                'Success',
                'success'
                'Succeeded',
                '',
                None,
            ),
        },
        'prompt': {
            'web': False,
        },
        'fetch_pipes_keys': {
            'negation_prefix': '_',
        },
    },
    'connectors': {
        'default_label': 'main',
    },
    'stack': {
        'dollar_standin': '<DOLLAR>', ### Temporary replacement for docker-compose.yaml.
    },
    'users': {
        'password_hash': {
            'schemes': [
                'pbkdf2_sha256',
            ],
            'default': 'pbkdf2_sha256',
            'pbkdf2_sha256__default_rounds': 30000,
        },
        'min_username_length': 1,
        'max_username_length': 26,
        'min_password_length': 5,
    },
    'plugins': {
        'repo_separator': '@',
        'lock_sleep_total': 1.0,
        'lock_sleep_increment': 0.1,
    },
    'pipes': {
        'dtypes': {
            'min_ratio_columns_changed_for_full_astype': 0.5,
        },
        'exists_timeout_seconds': 5.0,
    },
    'setup': {
        'name': 'meerschaum',
        'formal_name': 'Meerschaum',
        'app_id': 'io.meerschaum',
        'description': 'Sync Time-Series Pipes with Meerschaum',
        'url': 'https://meerschaum.io',
        'project_urls': {
            'Documentation': 'https://docs.meerschaum.io',
            'Changelog': 'https://meerschaum.io/news/changelog',
            'GitHub': 'https://github.com/bmeares/Meerschaum',
            'Homepage': 'https://meerschaum.io',
            'Donate': 'https://github.com/sponsors/bmeares',
            'Discord': 'https://discord.gg/8U8qMUjvcc',
        },
        'author': 'Bennett Meares',
        'author_email': 'bennett.meares@gmail.com',
        'maintainer_email': 'bennett.meares@gmail.com',
        'license': 'Apache Software License 2.0',
    },
}


def _static_config():
    """
    Alias function for the global `STATIC_CONFIG` dictionary.
    """
    return STATIC_CONFIG



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for building graphs of pipes' data.
"""

from __future__ import annotations
from meerschaum.utils.typing import WebState, List, Optional
from meerschaum.api import debug, CHECK_UPDATE
from meerschaum.utils.packages import attempt_import, import_dcc, import_html, import_pandas
#  import plotly.express as px
pd = import_pandas(check_update=CHECK_UPDATE)
px = attempt_import('plotly.express', warn=False, check_update=CHECK_UPDATE)
dbc = attempt_import('dash_bootstrap_components', check_update=CHECK_UPDATE)
html, dcc = import_html(check_update=CHECK_UPDATE), import_dcc(check_update=CHECK_UPDATE)

def get_graphs_cards(state: Optional[WebState]):
    """
    Return a tuple of lists containing cards and alerts.
    """
    from meerschaum.api.dash.pipes import pipes_from_state
    pipes = pipes_from_state(state, as_list=True)
    cards, alerts = [], []
    for pipe in pipes:
        dt_name, id_name = pipe.get_columns('datetime', 'id', error=False)
        val_name = pipe.get_val_column(debug=debug)
        if dt_name is not None and val_name is not None:
            df = pipe.get_backtrack_data(backtrack_minutes=1440)
            fig_args = {
                'data_frame': df,
                'x': dt_name,
                'y': val_name,
                'line_group': id_name,
                'title': f"{pipe}"
            }
            try:
                fig = px.line(**fig_args)
                graph = dcc.Graph(figure=fig)
                body = graph
            except Exception as e:
                body = html.P(
                    f"Unable to create graph for {pipe}. "
                    + "Please check that the columns are correctly set."
                )
        else:
            body = html.H4(f"Missing columns for {pipe}.")
        card = dbc.Card([
            #  dbc.CardHeader(html.H4(str(pipe))),
            dbc.CardBody(body),
            #  dbc.CardFooter(),
        ])
        cards.append(card)

    return cards, alerts



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Implement prompt_toolkit's AutoSuggest base class.
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional

from meerschaum.utils.packages import attempt_import
prompt_toolkit_auto_suggest, prompt_toolkit_document, prompt_toolkit_buffer = (
    attempt_import('prompt_toolkit.auto_suggest'),
    attempt_import('prompt_toolkit.document'),
    attempt_import('prompt_toolkit.buffer'),
)
from meerschaum._internal.entry import get_shell

class ValidAutoSuggest(prompt_toolkit_auto_suggest.AutoSuggest):
    def get_suggestion(
        self,
        buffer: prompt_toolkit_buffer.Buffer,
        document: prompt_toolkit_document.Document,
    ) -> Optional[prompt_toolkit_auto_suggest.Suggestion]:
        """Only return valid commands from history."""
        history = buffer.history
        text = document.text.rsplit("\n", 1)[-1]
        if not text.strip():
            return None

        for string in reversed(list(history.get_strings())):
            for line in reversed(string.splitlines()):
                if line.startswith(text) and 'do_' + line.split(' ')[0] in dir(get_shell()):
                    return prompt_toolkit_auto_suggest.Suggestion(line[len(text) :])




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define custom SQLAlchemy types.
"""

import uuid
from meerschaum.utils.packages import attempt_import
sqlalchemy, sqlalchemy_dialects_postgresql = attempt_import(
    'sqlalchemy', 'sqlalchemy.dialects.postgresql'
)

BINARY = sqlalchemy.types.BINARY
psqlUUID = sqlalchemy_dialects_postgresql.UUID

class UUID(sqlalchemy.types.TypeDecorator):
    """
    Platform-independent GUID type.
    
    Uses PostgreSQL's UUID type, otherwise uses
    BINARY(16), to store UUID.

    """
    impl = BINARY

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(psqlUUID())
        else:
            return dialect.type_descriptor(BINARY(16))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                if isinstance(value, bytes):
                    value = uuid.UUID(bytes=value)
                elif isinstance(value, int):
                    value = uuid.UUID(int=value)
                elif isinstance(value, str):
                    value = uuid.UUID(value)
        if dialect.name == 'postgresql':
            return str(value)
        else:
            return value.bytes

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        if dialect.name == 'postgresql':
            return uuid.UUID(value)
        else:
            return uuid.UUID(bytes=value)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Retrieve Pipes' data from instances.
"""

from __future__ import annotations
from datetime import datetime, timedelta
from meerschaum.utils.typing import Optional, Dict, Any, Union, Generator, List, Tuple
from meerschaum.config import get_config

def get_data(
        self,
        select_columns: Optional[List[str]] = None,
        omit_columns: Optional[List[str]] = None,
        begin: Union[datetime, int, None] = None,
        end: Union[datetime, int, None] = None,
        params: Optional[Dict[str, Any]] = None,
        as_iterator: bool = False,
        as_chunks: bool = False,
        as_dask: bool = False,
        chunk_interval: Union[timedelta, int, None] = None,
        fresh: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> Union['pd.DataFrame', Generator['pd.DataFrame'], None]:
    """
    Get a pipe's data from the instance connector.

    Parameters
    ----------
    select_columns: Optional[List[str]], default None
        If provided, only select these given columns.
        Otherwise select all available columns (i.e. `SELECT *`).

    omit_columns: Optional[List[str]], default None
        If provided, remove these columns from the selection.

    begin: Union[datetime, int, None], default None
        Lower bound datetime to begin searching for data (inclusive).
        Translates to a `WHERE` clause like `WHERE datetime >= begin`.
        Defaults to `None`.

    end: Union[datetime, int, None], default None
        Upper bound datetime to stop searching for data (inclusive).
        Translates to a `WHERE` clause like `WHERE datetime < end`.
        Defaults to `None`.

    params: Optional[Dict[str, Any]], default None
        Filter the retrieved data by a dictionary of parameters.
        See `meerschaum.utils.sql.build_where` for more details. 

    as_iterator: bool, default False
        If `True`, return a generator of chunks of pipe data.

    as_chunks: bool, default False
        Alias for `as_iterator`.

    as_dask: bool, default False
        If `True`, return a `dask.DataFrame`
        (which may be loaded into a Pandas DataFrame with `df.compute()`).

    chunk_interval: Union[timedelta, int, None], default None
        If `as_iterator`, then return chunks with `begin` and `end` separated by this interval.
        This may be set under `pipe.parameters['chunk_minutes']`.
        By default, use a timedelta of 1440 minutes (1 day).
        If `chunk_interval` is an integer and the `datetime` axis a timestamp,
        the use a timedelta with the number of minutes configured to this value.
        If the `datetime` axis is an integer, default to the configured chunksize.
        If `chunk_interval` is a `timedelta` and the `datetime` axis an integer,
        use the number of minutes in the `timedelta`.

    fresh: bool, default True
        If `True`, skip local cache and directly query the instance connector.
        Defaults to `True`.

    debug: bool, default False
        Verbosity toggle.
        Defaults to `False`.

    Returns
    -------
    A `pd.DataFrame` for the pipe's data corresponding to the provided parameters.

    """
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.venv import Venv
    from meerschaum.connectors import get_connector_plugin
    from meerschaum.utils.misc import iterate_chunks, items_str
    from meerschaum.utils.dtypes import to_pandas_dtype
    from meerschaum.utils.dataframe import add_missing_cols_to_df
    from meerschaum.utils.packages import attempt_import
    dd = attempt_import('dask.dataframe') if as_dask else None
    dask = attempt_import('dask') if as_dask else None

    if select_columns == '*':
        select_columns = None
    elif isinstance(select_columns, str):
        select_columns = [select_columns]

    if isinstance(omit_columns, str):
        omit_columns = [omit_columns]

    as_iterator = as_iterator or as_chunks

    if as_iterator or as_chunks:
        return self._get_data_as_iterator(
            select_columns = select_columns,
            omit_columns = omit_columns,
            begin = begin,
            end = end,
            params = params,
            chunk_interval = chunk_interval,
            fresh = fresh,
            debug = debug,
        )

    if as_dask:
        from multiprocessing.pool import ThreadPool
        dask_pool = ThreadPool(self.get_num_workers())
        dask.config.set(pool=dask_pool)
        chunk_interval = self.get_chunk_interval(chunk_interval, debug=debug)
        bounds = self.get_chunk_bounds(
            begin = begin,
            end = end,
            bounded = False,
            chunk_interval = chunk_interval,
            debug = debug,
        )
        dask_chunks = [
            dask.delayed(self.get_data)(
                select_columns = select_columns,
                omit_columns = omit_columns,
                begin = chunk_begin,
                end = chunk_end,
                params = params,
                chunk_interval = chunk_interval,
                fresh = fresh,
                debug = debug,
            )
            for (chunk_begin, chunk_end) in bounds
        ]
        dask_meta = {
            col: to_pandas_dtype(typ)
            for col, typ in self.dtypes.items()
        }
        return dd.from_delayed(dask_chunks, meta=dask_meta)

    if not self.exists(debug=debug):
        return None
       
    if self.cache_pipe is not None:
        if not fresh:
            _sync_cache_tuple = self.cache_pipe.sync(
                begin = begin,
                end = end,
                params = params,
                debug = debug,
                **kw
            )
            if not _sync_cache_tuple[0]:
                warn(f"Failed to sync cache for {self}:\n" + _sync_cache_tuple[1])
                fresh = True
            else: ### Successfully synced cache.
                return self.enforce_dtypes(
                    self.cache_pipe.get_data(
                        select_columns = select_columns,
                        omit_columns = omit_columns,
                        begin = begin,
                        end = end,
                        params = params,
                        debug = debug,
                        fresh = True,
                        **kw
                    ),
                    debug = debug,
                )

    with Venv(get_connector_plugin(self.instance_connector)):
        df = self.instance_connector.get_pipe_data(
            pipe = self,
            select_columns = select_columns,
            omit_columns = omit_columns,
            begin = begin,
            end = end,
            params = params,
            debug = debug,
            **kw
        )
        if df is None:
            return df

        if not select_columns:
            select_columns = [col for col in df.columns]

        cols_to_omit = [
            col
            for col in df.columns
            if (
                col in (omit_columns or [])
                or
                col not in (select_columns or [])
            )
        ]
        cols_to_add = [
            col
            for col in select_columns
            if col not in df.columns
        ]
        if cols_to_omit:
            warn(
                (
                    f"Received {len(cols_to_omit)} omitted column"
                    + ('s' if len(cols_to_omit) != 1 else '')
                    + f" for {self}. "
                    + "Consider adding `select_columns` and `omit_columns` support to "
                    + f"'{self.instance_connector.type}' connectors to improve performance."
                ),
                stack = False,
            )
            _cols_to_select = [col for col in df.columns if col not in cols_to_omit]
            df = df[_cols_to_select]

        if cols_to_add:
            warn(
                (
                    f"Specified columns {items_str(cols_to_add)} were not found on {self}. "
                    + "Adding these to the DataFrame as null columns."
                ),
                stack = False,
            )
            df = add_missing_cols_to_df(df, {col: 'string' for col in cols_to_add})

        return self.enforce_dtypes(df, debug=debug)


def _get_data_as_iterator(
        self,
        select_columns: Optional[List[str]] = None,
        omit_columns: Optional[List[str]] = None,
        begin: Optional[datetime] = None,
        end: Optional[datetime] = None,
        params: Optional[Dict[str, Any]] = None,
        chunk_interval: Union[timedelta, int, None] = None,
        fresh: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> Generator['pd.DataFrame']:
    """
    Return a pipe's data as a generator.
    """
    from meerschaum.utils.misc import round_time
    parse_begin = isinstance(begin, str)
    parse_end = isinstance(end, str)
    if parse_begin or parse_end:
        from meerschaum.utils.packages import attempt_import
        dateutil_parser = attempt_import('dateutil.parser')
    if parse_begin:
        begin = dateutil_parser.parse(begin)
    if parse_end:
        end = dateutil_parser.parse(end)

    if not self.exists(debug=debug):
        return

    _ = kw.pop('as_chunks', None)
    _ = kw.pop('as_iterator', None)
    min_dt = (
        begin
        if begin is not None
        else self.get_sync_time(round_down=False, newest=False, params=params, debug=debug)
    )
    max_dt = (
        end
        if end is not None
        else self.get_sync_time(round_down=False, newest=True, params=params, debug=debug)
    )

    ### We want to search just past the maximum value.
    if end is None:
        if isinstance(max_dt, int):
            max_dt += 1
        elif isinstance(max_dt, datetime):
            max_dt = round_time(max_dt + timedelta(minutes=1))

    chunk_interval = self.get_chunk_interval(chunk_interval, debug=debug)

    ### If we can't determine bounds
    ### or if chunk_interval exceeds the max,
    ### return a single chunk.
    if (
        (min_dt is None and max_dt is None)
        or
        (min_dt + chunk_interval) > max_dt
    ):
        yield self.get_data(
            select_columns = select_columns,
            omit_columns = omit_columns,
            begin = begin,
            end = end,
            params = params,
            fresh = fresh,
            debug = debug,
        )
        return

    chunk_bounds = self.get_chunk_bounds(
        begin = min_dt,
        end = max_dt,
        chunk_interval = chunk_interval,
        debug = debug,
    )

    for chunk_begin, chunk_end in chunk_bounds:
        chunk = self.get_data(
            select_columns = select_columns,
            omit_columns = omit_columns,
            begin = chunk_begin,
            end = chunk_end,
            params = params,
            fresh = fresh,
            debug = debug,
        )
        if len(chunk) > 0:
            yield chunk


def get_backtrack_data(
        self,
        backtrack_minutes: Optional[int] = None,
        begin: Union[datetime, int, None] = None,
        params: Optional[Dict[str, Any]] = None,
        fresh: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> Optional['pd.DataFrame']:
    """
    Get the most recent data from the instance connector as a Pandas DataFrame.

    Parameters
    ----------
    backtrack_minutes: Optional[int], default None
        How many minutes from `begin` to select from.
        If `None`, use `pipe.parameters['fetch']['backtrack_minutes']`.

    begin: Optional[datetime], default None
        The starting point to search for data.
        If begin is `None` (default), use the most recent observed datetime
        (AKA sync_time).

        ```
        E.g. begin = 02:00

        Search this region.           Ignore this, even if there's data.
        /  /  /  /  /  /  /  /  /  |
        -----|----------|----------|----------|----------|----------|
        00:00      01:00      02:00      03:00      04:00      05:00

        ```

    params: Optional[Dict[str, Any]], default None
        The standard Meerschaum `params` query dictionary.
        
        
    fresh: bool, default False
        If `True`, Ignore local cache and pull directly from the instance connector.
        Only comes into effect if a pipe was created with `cache=True`.

    debug: bool default False
        Verbosity toggle.

    Returns
    -------
    A `pd.DataFrame` for the pipe's data corresponding to the provided parameters. Backtrack data
    is a convenient way to get a pipe's data "backtracked" from the most recent datetime.
    """
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.venv import Venv
    from meerschaum.connectors import get_connector_plugin

    if not self.exists(debug=debug):
        return None

    backtrack_interval = self.get_backtrack_interval(debug=debug)
    if backtrack_minutes is None:
        backtrack_minutes = (
            (backtrack_interval.total_seconds() * 60)
            if isinstance(backtrack_interval, timedelta)
            else backtrack_interval
        )

    if self.cache_pipe is not None:
        if not fresh:
            _sync_cache_tuple = self.cache_pipe.sync(begin=begin, params=params, debug=debug, **kw)
            if not _sync_cache_tuple[0]:
                warn(f"Failed to sync cache for {self}:\n" + _sync_cache_tuple[1])
                fresh = True
            else: ### Successfully synced cache.
                return self.enforce_dtypes(
                    self.cache_pipe.get_backtrack_data(
                        fresh = True,
                        begin = begin,
                        backtrack_minutes = backtrack_minutes,
                        params = params,
                        debug = deubg,
                        **kw
                    ),
                    debug = debug,
                )

    if hasattr(self.instance_connector, 'get_backtrack_data'):
        with Venv(get_connector_plugin(self.instance_connector)):
            return self.enforce_dtypes(
                self.instance_connector.get_backtrack_data(
                    pipe = self,
                    begin = begin,
                    backtrack_minutes = backtrack_minutes,
                    params = params,
                    debug = debug,
                    **kw
                ),
                debug = debug,
            )

    if begin is None:
        begin = self.get_sync_time(params=params, debug=debug)

    backtrack_interval = (
        timedelta(minutes=backtrack_minutes)
        if isinstance(begin, datetime)
        else backtrack_minutes
    )
    if begin is not None:
        begin = begin - backtrack_interval

    return self.get_data(
        begin = begin,
        params = params,
        debug = debug,
        **kw
    )


def get_rowcount(
        self,
        begin: Optional[datetime] = None,
        end: Optional['datetime'] = None,
        params: Optional[Dict[str, Any]] = None,
        remote: bool = False,
        debug: bool = False
    ) -> int:
    """
    Get a Pipe's instance or remote rowcount.

    Parameters
    ----------
    begin: Optional[datetime], default None
        Count rows where datetime > begin.

    end: Optional[datetime], default None
        Count rows where datetime < end.

    remote: bool, default False
        Count rows from a pipe's remote source.
        **NOTE**: This is experimental!

    debug: bool, default False
        Verbosity toggle.

    Returns
    -------
    An `int` of the number of rows in the pipe corresponding to the provided parameters.
    Returned 0 if the pipe does not exist.
    """
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.venv import Venv
    from meerschaum.connectors import get_connector_plugin

    connector = self.instance_connector if not remote else self.connector
    try:
        with Venv(get_connector_plugin(connector)):
            rowcount = connector.get_pipe_rowcount(
                self,
                begin = begin,
                end = end,
                params = params,
                remote = remote,
                debug = debug,
            )
            if rowcount is None:
                return 0
            return rowcount
    except AttributeError as e:
        warn(e)
        if remote:
            return 0
    warn(f"Failed to get a rowcount for {self}.")
    return 0


def get_chunk_interval(
        self,
        chunk_interval: Union[timedelta, int, None] = None,
        debug: bool = False,
    ) -> Union[timedelta, int]:
    """
    Get the chunk interval to use for this pipe.

    Parameters
    ----------
    chunk_interval: Union[timedelta, int, None], default None
        If provided, coerce this value into the correct type.
        For example, if the datetime axis is an integer, then
        return the number of minutes.

    Returns
    -------
    The chunk interval (`timedelta` or `int`) to use with this pipe's `datetime` axis.
    """
    default_chunk_minutes = get_config('pipes', 'parameters', 'verify', 'chunk_minutes')
    configured_chunk_minutes = self.parameters.get('verify', {}).get('chunk_minutes', None)
    chunk_minutes = (
        (configured_chunk_minutes or default_chunk_minutes)
        if chunk_interval is None
        else (
            chunk_interval
            if isinstance(chunk_interval, int)
            else int(chunk_interval.total_seconds() / 60)
        )
    )

    dt_col = self.columns.get('datetime', None)
    if dt_col is None:
        return timedelta(minutes=chunk_minutes)

    dt_dtype = self.dtypes.get(dt_col, 'datetime64[ns]')
    if 'int' in dt_dtype.lower():
        return chunk_minutes
    return timedelta(minutes=chunk_minutes)


def get_chunk_bounds(
        self,
        begin: Union[datetime, int, None] = None,
        end: Union[datetime, int, None] = None,
        bounded: bool = False,
        chunk_interval: Union[timedelta, int, None] = None,
        debug: bool = False,
    ) -> List[
        Tuple[
            Union[datetime, int, None],
            Union[datetime, int, None],
        ]
    ]:
    """
    Return a list of datetime bounds for iterating over the pipe's `datetime` axis.

    Parameters
    ----------
    begin: Union[datetime, int, None], default None
        If provided, do not select less than this value.
        Otherwise the first chunk will be unbounded.

    end: Union[datetime, int, None], default None
        If provided, do not select greater than or equal to this value.
        Otherwise the last chunk will be unbounded.

    bounded: bool, default False
        If `True`, do not include `None` in the first chunk.

    chunk_interval: Union[timedelta, int, None], default None
        If provided, use this interval for the size of chunk boundaries.
        The default value for this pipe may be set
        under `pipe.parameters['verify']['chunk_minutes']`.

    debug: bool, default False
        Verbosity toggle.

    Returns
    -------
    A list of chunk bounds (datetimes or integers).
    If unbounded, the first and last chunks will include `None`.
    """
    include_less_than_begin = not bounded and begin is None
    include_greater_than_end = not bounded and end is None
    if begin is None:
        begin = self.get_sync_time(newest=False, debug=debug)
    if end is None:
        end = self.get_sync_time(newest=True, debug=debug)
    if begin is None and end is None:
        return [(None, None)]

    ### Set the chunk interval under `pipe.parameters['verify']['chunk_minutes']`.
    chunk_interval = self.get_chunk_interval(chunk_interval, debug=debug)
    
    ### Build a list of tuples containing the chunk boundaries
    ### so that we can sync multiple chunks in parallel.
    ### Run `verify pipes --workers 1` to sync chunks in series.
    chunk_bounds = []
    begin_cursor = begin
    while begin_cursor < end:
        end_cursor = begin_cursor + chunk_interval
        chunk_bounds.append((begin_cursor, end_cursor))
        begin_cursor = end_cursor

    ### The chunk interval might be too large.
    if not chunk_bounds and end >= begin:
        chunk_bounds = [(begin, end)]

    ### Truncate the last chunk to the end timestamp.
    if chunk_bounds[-1][1] > end:
        chunk_bounds[-1] = (chunk_bounds[-1][0], end)

    ### Pop the last chunk if its bounds are equal.
    if chunk_bounds[-1][0] == chunk_bounds[-1][1]:
        chunk_bounds = chunk_bounds[:-1]

    if include_less_than_begin:
        chunk_bounds = [(None, begin)] + chunk_bounds
    if include_greater_than_end:
        chunk_bounds = chunk_bounds + [(end, None)]

    return chunk_bounds



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
The default configuration values to write to config.yaml.
"""

import sys, os, multiprocessing

from meerschaum.connectors import attributes as connector_attributes
from meerschaum.config._paths import SQLITE_DB_PATH

default_meerschaum_config = {
    'instance': 'sql:main',
    'api_instance': 'MRSM{meerschaum:instance}',
    'web_instance': 'MRSM{meerschaum:instance}',
    'default_repository': 'api:mrsm',
    'connectors': {
        'sql': {
            'default': {},
            'main': {
                'username': 'mrsm',
                'password': 'mrsm',
                'flavor': 'timescaledb',
                'host': 'localhost',
                'database': 'meerschaum',
                'port': 5432,
            },
            'local': {
                'flavor': 'sqlite',
                'database': str(SQLITE_DB_PATH),
            },
            'memory': {
                'flavor': 'sqlite',
                'database': ':memory:',
            },
        },
        'api': {
            'default': connector_attributes['api']['default'],
            'main': {
                'host': 'localhost',
                'port': 8000,
            },
            'local': {
                'host': 'localhost',
            },
            'mrsm': {
                'host': 'api.mrsm.io',
                'port': 443,
                'protocol': 'https',
            },
        },
    },
}
default_system_config = {
    'connectors': {
        'all': {
            'pandas': 'pandas',
        },
        'sql': {
            'instance': {
                'stale_temporary_tables_minutes': 1440,
            },
            'chunksize': 100_000,
            'poolclass': 'sqlalchemy.pool.QueuePool',
            'create_engine': {
                'method': 'multi',
                'pool_size': 8,
                'max_overflow': 12,
                'pool_recycle': 3600,
                'connect_args': {},
            },
        },

        'api' : {
        },
    },
    ### not to be confused with system_config['connectors']['api'], this is the configuration
    ### for the API server itself.
    'api': {
        'uvicorn': {
            'app': 'meerschaum.api:app',
            'port': 8000,
            'host': '0.0.0.0',
            'workers': max(int(multiprocessing.cpu_count() / 2), 1),
            'proxy_headers': True,
            'forwarded_allow_ips': '*',
        },
        'permissions':       {
            'registration': {
                'users': True,
                'pipes': True,
                'plugins': True,
            },
            'actions': {
                'non_admin': True,
            },
            'chaining' : {
                'insecure_parent_instance': False,
                'child_apis': False,
            },
        },
        'protocol': default_meerschaum_config['connectors']['api']['default']['protocol'],
    },
    'experimental': {
        'fetch': False,
        'cache': True,
        'space': False,
        'join_fetch': False,
        'inplace_sync': True,
    },
}
default_pipes_config = {
    'parameters': {
        'columns': {
            'datetime': None,
            'id': None,
        },
        'fetch': {
            'backtrack_minutes': 1440,
        },
        'verify': {
            'bound_days': 366,
            'chunk_minutes': 1440,
        },
    },
    'attributes': {
        'local_cache_timeout_seconds': 60,
    },
    'sync': {
        'filter_params_index_limit': 250,
    },
}
default_plugins_config = {}
default_experimental_config = {
    'venv': True,
}



### build default config dictionary
default_config = {}
default_config['meerschaum'] = default_meerschaum_config
default_config['system'] = default_system_config
from meerschaum.config._formatting import default_formatting_config
default_config['formatting'] = default_formatting_config
from meerschaum.config._shell import default_shell_config
default_config['shell'] = default_shell_config
default_config['pipes'] = default_pipes_config
default_config['plugins'] = default_plugins_config
from meerschaum.config._jobs import default_jobs_config
default_config['jobs'] = default_jobs_config
#  default_config['experimental'] = default_experimental_config
### add configs from other packages
try:
    import meerschaum.config.stack
except ImportError as e:
    pass
finally:
    from meerschaum.config.stack import default_stack_config
default_config['stack'] = default_stack_config

default_header_comment = """
#####################################################################
#                                                                   #
#  Edit or add credentials for connectors.                          #
#  You can read more about connectors at https://meerschaum.io.     #
#                                                                   #
#  Connectors inherit from `default`, and flavor-dependent defaults #
#  are defined for SQL connectors (e.g. port 5432 for PostgreSQL).  #
#                                                                   #
#####################################################################

"""



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Utilities for formatting output text
"""

from __future__ import annotations
import platform
import os
import sys
from meerschaum.utils.typing import Optional, Union, Any
from meerschaum.utils.formatting._shell import make_header
from meerschaum.utils.formatting._pprint import pprint
from meerschaum.utils.formatting._pipes import (
    pprint_pipes,
    highlight_pipes,
    format_pipe_success_tuple,
    print_pipes_results,
    extract_stats_from_message,
    pipe_repr,
)
from meerschaum.utils.threading import Lock, RLock

_attrs = {
    'ANSI': None,
    'UNICODE': None,
    'CHARSET': None,
}
__all__ = sorted([
    'ANSI', 'CHARSET', 'UNICODE',
    'colored',
    'translate_rich_to_termcolor',
    'get_console',
    'print_tuple',
    'print_options',
    'fill_ansi',
    'pprint',
    'highlight_pipes',
    'pprint_pipes',
    'make_header',
    'pipe_repr',
    'print_pipes_results',
    'extract_stats_from_message',
])
__pdoc__ = {}
_locks = {
    '_colorama_init': RLock(),
}


def colored_fallback(*args, **kw):
    return ' '.join(args)

def translate_rich_to_termcolor(*colors) -> tuple:
    """Translate between rich and more_termcolor terminology.
    This is probably prone to breaking.

    Parameters
    ----------
    *colors :
        

    Returns
    -------

    """
    _colors = []
    for c in colors:
        _c_list = []
        ### handle 'bright'
        c = c.replace('bright_', 'bright ')

        ### handle 'on'
        if ' on ' in c:
            _on = c.split(' on ')
            _colors.append(_on[0])
            for _c in _on[1:]:
                _c_list.append('on ' + _c)
        else:
            _c_list += [c]

        _colors += _c_list

    return tuple(_colors)


def rich_text_to_str(text: 'rich.text.Text') -> str:
    """Convert a `rich.text.Text` object to a string with ANSI in-tact."""
    _console = get_console()
    if _console is None:
        return str(text)
    with console.capture() as cap:
        console.print(text)
    string = cap.get()
    return string[:-1]


def _init():
    """
    Initial color settings (mostly for Windows).
    """
    if platform.system() != "Windows":
        return
    if 'PYTHONIOENCODING' not in os.environ:
        os.environ['PYTHONIOENCODING'] = 'utf-8'
    if 'PYTHONLEGACYWINDOWSSTDIO' not in os.environ:
        os.environ['PYTHONLEGACYWINDOWSSTDIO'] = 'utf-8'
    sys.stdin.reconfigure(encoding='utf-8')
    sys.stdout.reconfigure(encoding='utf-8')
    sys.stderr.reconfigure(encoding='utf-8')

    from ctypes import windll
    k = windll.kernel32
    k.SetConsoleMode(k.GetStdHandle(-11), 7)
    os.system("color")

    from meerschaum.utils.packages import attempt_import
    ### init colorama for Windows color output
    colorama, more_termcolor = attempt_import(
        'colorama',
        'more_termcolor',
        lazy = False,
        warn = False,
        color = False,
    )
    try:
        colorama.init(autoreset=False)
        success = True
    except Exception as e:
        import traceback
        traceback.print_exc()
        _attrs['ANSI'], _attrs['UNICODE'], _attrs['CHARSET'] = False, False, 'ascii'
        success = False

    if more_termcolor is None:
        _attrs['ANSI'], _attrs['UNICODE'], _attrs['CHARSET'] = False, False, 'ascii'
        success = False

    return success

_colorama_init = False
def colored(text: str, *colors, as_rich_text: bool=False, **kw) -> Union[str, 'rich.text.Text']:
    """Apply colors and rich styles to a string.
    If a `style` keyword is provided, a `rich.text.Text` object will be parsed into a string.
    Otherwise attempt to use the legacy `more_termcolor.colored` method.

    Parameters
    ----------
    text: str
        The string to apply formatting to.
        
    *colors:
        A list of colors to pass to `more_termcolor.colored()`.
        Has no effect if `style` is provided.

    style: str, default None
        If provided, pass to `rich` for processing.

    as_rich_text: bool, default False
        If `True`, return a `rich.Text` object.
        `style` must be provided.
        
    **kw:
        Keyword arguments to pass to `rich.text.Text` or `more_termcolor`.
        

    Returns
    -------
    An ANSI-formatted string or a `rich.text.Text` object if `as_rich_text` is `True`.

    """
    from meerschaum.utils.packages import import_rich, attempt_import
    global _colorama_init
    _init()
    with _locks['_colorama_init']:
        if not _colorama_init:
            _colorama_init = _init()

    if 'style' in kw:
        rich = import_rich()
        rich_text = attempt_import('rich.text')
        text_obj = rich_text.Text(text, **kw)
        if as_rich_text:
            return text_obj
        return rich_text_to_str(text_obj)

    more_termcolor = attempt_import('more_termcolor', lazy=False)
    try:
        colored_text = more_termcolor.colored(text, *colors, **kw)
    except Exception as e:
        colored_text = None

    if colored_text is not None:
        return colored_text

    try:
        _colors = translate_rich_to_termcolor(*colors)
        colored_text = more_termcolor.colored(text, *_colors, **kw)
    except Exception as e:
        colored_text = None

    if colored_text is None:
        ### NOTE: warn here?
        return text

    return colored_text

console = None
def get_console():
    """Get the rich console."""
    global console
    from meerschaum.utils.packages import import_rich, attempt_import
    rich = import_rich()
    rich_console = attempt_import('rich.console')
    try:
        console = rich_console.Console(force_terminal=True, color_system='truecolor')
    except Exception as e:
        console = None
    return console


def print_tuple(
        tup: tuple,
        skip_common: bool = True,
        common_only: bool = False,
        upper_padding: int = 0,
        lower_padding: int = 0,
        calm: bool = False,
        _progress: Optional['rich.progress.Progress'] = None,
    ) -> None:
    """
    Print `meerschaum.utils.typing.SuccessTuple`.

    Parameters
    ----------
    skip_common: bool, default True
        If `True`, do not print common success tuples (i.e. `(True, "Success")`).

    common_only: bool, default False
        If `True`, only print if the success tuple is common.

    upper_padding: int, default 0
        How many newlines to prepend to the message.

    lower_padding: int, default 0
        How many newlines to append to the message.

    calm: bool, default False
        If `True`, use the default emoji and color scheme.
    """
    from meerschaum.config.static import STATIC_CONFIG
    _init()
    try:
        status = 'success' if tup[0] else 'failure'
    except TypeError:
        status = 'failure'
        tup = None, None

    if calm:
        status += '_calm'

    omit_messages = STATIC_CONFIG['system']['success']['ignore']

    do_print = True

    if common_only:
        skip_common = False
        do_print = tup[1] in omit_messages

    if skip_common:
        do_print = tup[1] not in omit_messages

    if do_print:
        ANSI, CHARSET = __getattr__('ANSI'), __getattr__('CHARSET')
        from meerschaum.config import get_config
        status_config = get_config('formatting', status, patch=True)

        msg = ' ' + status_config[CHARSET]['icon'] + ' ' + str(tup[1])
        lines = msg.split('\n')
        lines = [lines[0]] + [
            (('    ' + line if not line.startswith(' ') else line))
            for line in lines[1:]
        ]
        if ANSI:
            lines[0] = fill_ansi(highlight_pipes(lines[0]), **status_config['ansi']['rich'])
        msg = '\n'.join(lines)
        msg = ('\n' * upper_padding) + msg + ('\n' * lower_padding)
        print(msg)


def print_options(
        options: Optional[Dict[str, Any]] = None,
        nopretty: bool = False,
        no_rich: bool = False,
        name: str = 'options',
        header: Optional[str] = None,
        num_cols: Optional[int] = None,
        adjust_cols: bool = True,
        **kw
    ) -> None:
    """
    Print items in an iterable as a fancy table.

    Parameters
    ----------
    options: Optional[Dict[str, Any]], default None
        The iterable to be printed.

    nopretty: bool, default False
        If `True`, don't use fancy formatting.

    no_rich: bool, default False
        If `True`, don't use `rich` to format the output.

    name: str, default 'options'
        The text in the default header after `'Available'`.

    header: Optional[str], default None
        If provided, override `name` and use this as the header text.

    num_cols: Optional[int], default None
        How many columns in the table. Depends on the terminal size. If `None`, use 8.

    adjust_cols: bool, default True
        If `True`, adjust the number of columns depending on the terminal size.

    """
    import os
    from meerschaum.utils.packages import import_rich
    from meerschaum.utils.formatting import make_header, highlight_pipes
    from meerschaum.actions import actions as _actions
    from meerschaum.utils.misc import get_cols_lines, string_width, iterate_chunks


    if options is None:
        options = {}
    _options = []
    for o in options:
        _options.append(str(o))
    _header = f"Available {name}" if header is None else header

    if num_cols is None:
        num_cols = 8

    def _print_options_no_rich():
        if not nopretty:
            print()
            print(make_header(_header))
        ### print actions
        for option in sorted(_options):
            if not nopretty:
                print("  - ", end="")
            print(option)
        if not nopretty:
            print()

    rich = import_rich()
    if rich is None or nopretty or no_rich:
        _print_options_no_rich()
        return None

    ### Prevent too many options from being truncated on small terminals.
    if adjust_cols and _options:
        _cols, _lines = get_cols_lines()
        while num_cols > 1:
            cell_len = int(((_cols - 4) - (3 * (num_cols - 1))) / num_cols)
            num_too_big = sum([(1 if string_width(o) > cell_len else 0) for o in _options])
            if num_too_big > int(len(_options) / 3):
                num_cols -= 1
                continue
            break

    from meerschaum.utils.formatting import pprint, get_console
    from meerschaum.utils.packages import attempt_import
    rich_columns = attempt_import('rich.columns')
    rich_panel = attempt_import('rich.panel')
    rich_table = attempt_import('rich.table')
    Text = attempt_import('rich.text').Text
    box = attempt_import('rich.box')
    Panel = rich_panel.Panel
    Columns = rich_columns.Columns
    Table = rich_table.Table

    if _header is not None:
        table = Table(
            title = '\n' + _header,
            box = box.SIMPLE,
            show_header = False,
            show_footer = False,
            title_style = '',
            expand = True,
        )
    else:
        table = Table.grid(padding=(0, 2))
    for i in range(num_cols):
        table.add_column()

    chunks = iterate_chunks(
        [Text.from_ansi(highlight_pipes(o)) for o in sorted(_options)],
        num_cols,
        fillvalue=''
    )
    for c in chunks:
        table.add_row(*c)

    get_console().print(table)
    return None


def fill_ansi(string: str, style: str = '') -> str:
    """
    Fill in non-formatted segments of ANSI text.

    Parameters
    ----------
    string: str
        A string which contains ANSI escape codes.

    style: str
        Style arguments to pass to `rich.text.Text`.

    Returns
    -------
    A string with ANSI styling applied to the segments which don't yet have a style applied.
    """
    from meerschaum.utils.packages import import_rich, attempt_import
    from meerschaum.utils.misc import iterate_chunks
    rich = import_rich()
    Text = attempt_import('rich.text').Text
    try:
        msg = Text.from_ansi(string)
    except AttributeError as e:
        import traceback
        traceback.print_stack()
        msg = ''
    plain_indices = []
    for left_span, right_span in iterate_chunks(msg.spans, 2, fillvalue=len(msg)):
        left = left_span.end
        right = right_span.start if not isinstance(right_span, int) else right_span
        if left != right:
            plain_indices.append((left, right))
    if msg.spans:
        if msg.spans[0].start != 0:
            plain_indices = [(0, msg.spans[0].start)] + plain_indices
        if plain_indices and msg.spans[-1].end != len(msg) and plain_indices[-1][1] != len(msg):
            plain_indices.append((msg.spans[-1].end, len(msg)))

    if plain_indices:
        for left, right in plain_indices:
            msg.stylize(style, left, right)
    else:
        msg = Text(str(msg), style)

    return rich_text_to_str(msg)

def __getattr__(name: str) -> str:
    """
    Lazily load module-level variables.
    """
    if name.startswith('__') and name.endswith('__'):
        raise AttributeError("Cannot import dunders from this module.")

    if name in _attrs:
        if _attrs[name] is not None:
            return _attrs[name]
        from meerschaum.config import get_config
        if name.lower() in get_config('formatting'):
            _attrs[name] = get_config('formatting', name.lower())
        elif name == 'CHARSET':
            _attrs[name] = 'unicode' if __getattr__('UNICODE') else 'ascii'
        return _attrs[name]
    
    if name == '__wrapped__':
        import sys
        return sys.modules[__name__]
    if name == '__all__':
        return __all__

    try:
        return globals()[name]
    except KeyError:
        raise AttributeError(f"Could not find '{name}'")



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Meerschaum wrapper around YAML libraries.

This is so switching between PyYAML and ruamel.yaml is smoother.
"""

from meerschaum.utils.misc import filter_keywords
from meerschaum.utils.packages import attempt_import, all_packages, _import_module
from meerschaum.utils.warnings import error, warn
from meerschaum.utils.threading import Lock

_lib = None
### Also supports 'ruamel.yaml'.
_import_name = 'yaml'

_yaml = None
_locks = {
    '_lib': Lock(),
    '_yaml': Lock(),
}

__pdoc__ = {
    'attempt_import': False,
    'error': False,
    'all_packages': False,
}


def _string_presenter(dumper, data: str):
    """
    Format strings with newlines as blocks.
    https://stackoverflow.com/a/33300001/9699829
    """
    tag_str = 'tag:yaml.org,2002:str'
    kw = {}
    if len(data.splitlines()) > 1:
        kw['style'] = '|'
    return dumper.represent_scalar(tag_str, data, **kw)


class yaml:
    """
    Wrapper around `PyYAML` and `ruamel.yaml` so that we may switch between implementations.
    """
    global _yaml, _lib
    if _import_name is None:
        error(f"No YAML library declared.")
    with _locks['_lib']:
        try:
            _lib = _import_module(_import_name)
        except (ImportError, ModuleNotFoundError):
            _lib = attempt_import(_import_name, split=False, lazy=False, install=True)
    with _locks['_yaml']:
        _yaml = _lib if _import_name != 'ruamel.yaml' else _lib.YAML()
        if _import_name != 'ruamel.yaml':
            _yaml.add_representer(str, _string_presenter)
            _yaml.representer.SafeRepresenter.add_representer(str, _string_presenter)


    @staticmethod
    def safe_load(*args, **kw):
        """
        Execute `safe_load` for `PyYAML` and `load` for `ruamel.yaml`.
        """
        if _import_name == 'ruamel.yaml':
            return _yaml.load(*args, **filter_keywords(_yaml.load, **kw))
        return _yaml.safe_load(*args, **filter_keywords(_yaml.safe_load, **kw))


    @staticmethod
    def load(*args, **kw):
        """
        Execute `yaml.load()`.
        Handles the breaking change at `v6.0` of `PyYAML`
        (added `yaml.Loader` as a positional argument).
        """
        packaging_version = attempt_import('packaging.version')
        _args = list(args)
        if (
            _import_name == 'yaml'
            and packaging_version.parse(_yaml.__version__) >= packaging_version.parse('6.0')
        ):
            _args += [_yaml.Loader]
        return _yaml.load(*_args, **filter_keywords(_yaml.load, **kw))


    @staticmethod
    def dump(data, stream=None, **kw):
        """
        Dump to a stream. If no stream is provided, return a string instead.
        For `ruamel.yaml`, it dumps into a `StringIO` stream and returns `getvalue()`.
        """
        get_string = False
        if stream is None and _import_name == 'ruamel.yaml':
            stream = _lib.compat.StringIO()
            get_string = True
        result = _yaml.dump(data, stream, **filter_keywords(_yaml.dump, **kw))
        if get_string:
            return stream.getvalue()
        return result



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Import all routes from other modules in package
"""

from meerschaum.api.routes._login import *
from meerschaum.api.routes._actions import *
from meerschaum.api.routes._connectors import *
from meerschaum.api.routes._index import *
from meerschaum.api.routes._misc import *
from meerschaum.api.routes._pipes import *
from meerschaum.api.routes._plugins import *
from meerschaum.api.routes._users import *
from meerschaum.api.routes._version import *
from meerschaum.api.routes._webterm import *



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

import setuptools, sys, platform
from setuptools import setup
from setuptools.command.install import install
cx_Freeze, Executable = None, None

### read values from within the package
exec(open('meerschaum/config/_version.py').read())
exec(open('meerschaum/utils/packages/_packages.py').read())
exec(open('meerschaum/config/static/__init__.py').read())
setup_cf = _static_config()['setup']

class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        from meerschaum.actions import actions
        try:
            actions['verify'](['packages'], debug=False)
        except Exception as e:
            pass

extras = {}
### NOTE: package dependencies are defined in meerschaum.utils.packages._packages
for group in packages:
    extras[group] = [ install_name for import_name, install_name in packages[group].items() ]

with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

setup_kw_args = {
    'name'                          : setup_cf['name'],
    'version'                       : __version__,
    'description'                   : setup_cf['description'],
    'long_description'              : readme,
    'long_description_content_type' : 'text/markdown',
    'url'                           : setup_cf['url'],
    'project_urls'                  : setup_cf['project_urls'],
    'author'                        : setup_cf['author'],
    'author_email'                  : setup_cf['author_email'],
    'maintainer_email'              : setup_cf['maintainer_email'],
    'license'                       : setup_cf['license'],
    'packages'                      : setuptools.find_packages(exclude=['*test*']),
    'install_requires'              : extras['required'],
    'extras_require'                : extras,
    'entry_points'                  : {
        'console_scripts'           : [
            'mrsm = meerschaum.__main__:main',
            'meerschaum = meerschaum.__main__:main',
        ],
    },
    'cmdclass'                      : {
        'install'                   : PostInstallCommand,
    },
    'zip_safe'                      : True,
    'package_data'                  : {'' : ['*.html', '*.css', '*.js', '*.png', '*.ico']},
    'python_requires'               : '>=3.8',
    'classifiers'                   : [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Programming Language :: SQL",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Database",
        "Natural Language :: English",
    ],
}

setup(**setup_kw_args)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Misc API routes
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional

def get_mrsm_version(self, **kw) -> Optional[str]:
    """
    Return the Meerschaum version of the API instance.
    """
    from meerschaum.config.static import STATIC_CONFIG
    try:
        j = self.get(
            STATIC_CONFIG['api']['endpoints']['version'] + '/mrsm',
            use_token = True,
            **kw
        ).json()
    except Exception as e:
        return None
    if isinstance(j, dict) and 'detail' in j:
        return None
    return j

def get_chaining_status(self, **kw) -> Optional[bool]:
    """
    Fetch the chaining status of the API instance.
    """
    from meerschaum.config.static import STATIC_CONFIG
    try:
        response = self.get(
            STATIC_CONFIG['api']['endpoints']['chaining'],
            use_token = True,
            **kw
        )
        if not response:
            return None
    except Exception as e:
        return None

    return response.json()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Implement WebSockets for the Meerschaum API via FastAPI.
"""

import time, uuid
from datetime import datetime, timezone
from meerschaum.api import (
    app, get_api_connector, get_uvicorn_config, debug, fastapi, endpoints
)
from meerschaum.api.dash.users import is_session_authenticated
from meerschaum.utils.typing import Optional

_websocket_endpoint = endpoints['websocket']

websockets = {}
sessions = {}

@app.websocket('/dashws')
@app.websocket('/dash/ws')
@app.websocket(_websocket_endpoint)
async def websocket_endpoint(
        websocket: fastapi.WebSocket,
    ):
    """
    Communicate with the Web Interface over a websocket.
    """
    await websocket.accept()
    try:
        initial_data = await websocket.receive_json()
    except Exception as e:
        initial_data = {'session-id': None}
    session_id = initial_data.get('session-id', None)
    if not is_session_authenticated(str(session_id)):
        await websocket.close()
        return
    now = datetime.now(timezone.utc)
    join_msg = str(now)
    await websocket.send_text(join_msg)
    websockets[session_id] = websocket
    while True:
        try:
            data = await websocket.receive_text()
            await websocket.send_text(str(now))
        except fastapi.WebSocketDisconnect:
            delete_websocket_session(session_id)
            break


def delete_websocket_session(session_id: str) -> None:
    """
    Delete a websocket session if it exists.
    """
    try:
        del websockets[session_id]
    except KeyError:
        pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Formatting functions for the interactive shell
"""

from re import sub
from meerschaum.utils.threading import Lock
_locks = {'_tried_clear_command': Lock()}

def make_header(
        message : str,
        ruler : str = '─',
    ) -> str:
    """Format a message string with a ruler.
    Length of the ruler is the length of the longest word.
    
    Example:
        'My\nheader' -> 'My\nheader\n──────'
    """

    from meerschaum.utils.formatting import ANSI, UNICODE, colored
    if not UNICODE:
        ruler = '-'
    words = message.split('\n')
    max_length = 0
    for w in words:
        length = len(w)
        if length > max_length:
            max_length = length

    s = message + "\n"
    for i in range(max_length):
        s += ruler
    return s


_tried_clear_command = None
def clear_screen(debug: bool = False) -> bool:
    """Clear the terminal window of all text. If ANSI is enabled,
    print the ANSI code for clearing. Otherwise, execute `clear` or `cls`.
    """
    import platform
    import os
    from meerschaum.utils.formatting import ANSI, get_console
    from meerschaum.utils.debug import dprint
    from meerschaum.config import get_config
    global _tried_clear_command
    if not get_config('shell', 'clear_screen'):
        return True
    print("", end="", flush=True)
    if debug:
        dprint("Skipping screen clear.")
        return True
    if ANSI and platform.system() != 'Windows':
        if get_console() is not None:
            get_console().clear()
            print("", end="", flush=True)
            return True
        clear_string = (
            '\033c' if platform.system() != 'Windows'
            else '\033[2J'
        )
        reset_string = '\033[0m'
        print(clear_string + reset_string, end="")
        print("", end="", flush=True)
        return True
    ### ANSI support is disabled, try system level instead
    if _tried_clear_command is not None:
        return os.system(_tried_clear_command) == 0
    
    rc = -1
    command = 'clear' if platform.system() != 'Windows' else 'cls'
    try:
        rc = os.system(command)
    except Exception as e:
        pass
    if rc == 0:
        with _locks['_tried_clear_command']:
            _tried_clear_command = command
    return rc == 0


def flush_with_newlines(debug: bool = False) -> None:
    """Print newlines such that the entire terminal is cleared and new text will show up at the bottom."""
    import sys
    from meerschaum.utils.misc import get_cols_lines
    from meerschaum.utils.debug import dprint
    if debug:
        dprint("Skipping screen clear.")
        return
    cols, lines = get_cols_lines()
    sys.stderr.write('\n' * lines)


def progress(transient: bool = True, **kw):
    """
    Return the shell's `Progress` object.
    """
    from meerschaum.utils.packages import import_rich, attempt_import
    from meerschaum.utils.formatting import get_console, _init
    _init()
    rich = import_rich()
    rich_progress = attempt_import('rich.progress')
    return rich_progress.Progress(
        rich_progress.TextColumn(''),
        rich_progress.SpinnerColumn('clock'),
        rich_progress.TimeElapsedColumn(),
        rich_progress.TextColumn(''),
        rich_progress.BarColumn(bar_width=None,),
        transient = transient,
        console = get_console(),
        **kw
    )

def live(**kw):
    """
    Return the shell's `Live` object.
    """
    from meerschaum.utils.packages import import_rich, attempt_import
    rich = import_rich()
    rich_live = attempt_import('rich.live')
    return rich_live.Live(**kw)




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Generate random names for jobs.
"""

from __future__ import annotations
import os, random
from meerschaum.utils.typing import Dict, List, Tuple
from meerschaum.config._paths import DAEMON_RESOURCES_PATH

_bank: Dict[str, Dict[str, List[str]]] = {
    'adjectives': {
        'colors': [
            'red', 'blue', 'green', 'orange', 'pink', 'amber',
            'bright', 'dark', 'neon',
        ],
        'sizes': [
            'big', 'small', 'large', 'huge', 'tiny', 'long', 'short', 'grand', 'mini', 'micro'
        ],
        'personalities': [
            'groovy', 'cool', 'awesome', 'nice', 'fantastic', 'sweet', 'great', 'amazing',
            'terrific', 'funky', 'fancy', 'sneaky', 'elegant', 'dreamy',
        ],
        'emotions': [
            'angry', 'happy', 'excited', 'suspicious', 'sad', 'thankful', 'grateful', 'satisfied',
        ],
        'sensations': [
            'sleepy', 'awake', 'alert', 'thirsty', 'comfy', 'warm', 'cold', 'chilly', 'soft',
            'smooth', 'chunky',
        ],
        'materials': [
            'golden', 'silver', 'metal', 'plastic', 'wool', 'wooden', 'nylon', 'fuzzy', 'silky',
        ],
        'qualities': [
            'expensive', 'cheap', 'premier', 'best', 'favorite', 'better', 'good', 'affordable',
        ],
    },
    'nouns' : {
        'animals': [
            'mouse', 'fox', 'horse', 'dragon', 'pig', 'hippo', 'elephant' , 'tiger', 'deer',
            'gerbil', 'snake', 'turtle', 'rhino', 'dog', 'cat', 'giraffe', 'rabbit', 'squirrel',
            'unicorn', 'lizard', 'lion', 'bear', 'gazelle', 'whale', 'dolphin', 'fish', 'butterfly',
            'ladybug', 'fly', 'shrimp', 'flamingo', 'parrot', 'tuna', 'panda', 'lemur', 'duck',
            'seal', 'walrus', 'seagull', 'iguana', 'salamander', 'kitten', 'puppy', 'octopus',
        ],
        'plants': [
            'tree', 'flower', 'vine', 'fern', 'palm', 'palmetto', 'oak', 'pine', 'rose', 'lily',
            'ivy',
        ],
        'foods': [
            'pizza', 'sushi', 'apple', 'banana', 'sandwich', 'burger', 'taco', 'bratwurst',
            'grape', 'coconut', 'bread', 'toast',
        ],
        'geographies': [
            'ocean', 'mountain', 'desert', 'forest', 'tundra', 'savanna', 'grassland', 'prairie',
            'lake', 'city', 'river',
        ],
        'vehicles': [
            'car', 'bike', 'boat', 'bus', 'trolley', 'tram', 'plane', 'skates', 'skateboard',
            'kayak', 'canoe', 'paddleboard', 'skis', 'snowboard', 'truck', 'bicycle', 'unicycle',
            'tricycle',
        ],
        'clothing': [
            'socks', 'shirt', 'dress', 'shoes', 'hat', 'glasses', 'pocket', 'shorts', 'pants',
            'skirt', 'capris', 'helmet',
        ],
        'instruments': [
            'guitar', 'piano', 'trombone', 'drums', 'viola', 'violin', 'trumpet', 'bass',
            'harmonica', 'banjo',
        ],
    },
}

_disallow_combinations: List[Tuple[str, str]] = []

_adjectives: List[str]= []
for category, items in _bank['adjectives'].items():
    _adjectives += items

_nouns: List[str] = []
for category, items in _bank['nouns'].items():
    _nouns += items

def generate_random_name(separator: str = '_'):
    """
    Return a random adjective and noun combination.

    Parameters
    ----------
    separator: str, default '_'

    Returns
    -------
    A string containing an random adjective and random noun. 
    """
    while True:
        adjective_category = random.choice(list(_bank['adjectives'].keys()))
        noun_category = random.choice(list(_bank['nouns'].keys()))
        if (adjective_category, noun_category) in _disallow_combinations:
            continue
        break
    return (
        random.choice(_bank['adjectives'][adjective_category])
        + separator
        + random.choice(_bank['nouns'][noun_category])
    )


def get_new_daemon_name() -> str:
    """
    Generate a new random name until a unique one is found
    (up to ~6000 maximum possibilities).
    """
    existing_names = (
        os.listdir(DAEMON_RESOURCES_PATH)
        if DAEMON_RESOURCES_PATH.exists()
        else []
    )
    while True:
        _name = generate_random_name()
        if _name in existing_names:
            continue
        break
    return _name



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Return version information
"""

import fastapi
from meerschaum.api import app, endpoints, private, manager
from meerschaum.utils.typing import Union

@app.get(endpoints['version'], tags=['Version'])
def get_api_version(
        curr_user = (
            fastapi.Depends(manager) if private else None
        ),
    ):
    """
    Get the Meerschaum API version.
    """
    from meerschaum.api import __version__ as version
    return version

@app.get(endpoints['version'] + "/mrsm", tags=['Version'])
def get_meerschaum_version(
        curr_user = (
            fastapi.Depends(manager) if private else None
        ),
    ):
    """
    Get the Meerschaum instance version.
    """
    from meerschaum import __version__ as version
    return version




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Callbacks for jobs' cards.
"""

from __future__ import annotations
import json
import functools
import time
import traceback
import meerschaum as mrsm
from meerschaum.utils.typing import Optional, Dict, Any
from meerschaum.api import get_api_connector, endpoints, CHECK_UPDATE
from meerschaum.api.dash import dash_app, debug, active_sessions
from meerschaum.utils.packages import attempt_import, import_dcc, import_html
dash = attempt_import('dash', lazy=False, check_update=CHECK_UPDATE)
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State, ALL, MATCH
from dash import Patch
html, dcc = import_html(check_update=CHECK_UPDATE), import_dcc(check_update=CHECK_UPDATE)
import dash_bootstrap_components as dbc
from meerschaum.api.dash.components import alert_from_success_tuple, build_cards_grid
from meerschaum.utils.daemon import Daemon
from dash.exceptions import PreventUpdate
from meerschaum.api.dash.jobs import (
    build_manage_job_buttons_div_children,
    build_status_children,
    build_process_timestamps_children,
)
from meerschaum.api.dash.users import is_session_authenticated

@dash_app.callback(
    Output("download-logs", "data"),
    Input({'type': 'job-download-logs-button', 'index': ALL}, 'n_clicks'),
    prevent_initial_call = True,
)
def download_job_logs(n_clicks):
    """
    When the download logs button is clicked, download the logs as one text file.

    It would have been more efficient to return the existing files on disk,
    but because the rotating log will keep file sizes down, this should not grow too large.
    """
    if not n_clicks:
        raise PreventUpdate

    ctx = dash.callback_context.triggered
    if ctx[0]['value'] is None:
        raise PreventUpdate

    component_dict = json.loads(ctx[0]['prop_id'].split('.' + 'n_clicks')[0])
    daemon_id = component_dict['index']
    daemon = Daemon(daemon_id=daemon_id)
    return {
        'content': daemon.log_text,
        'filename': daemon.rotating_log.file_path.name,
    }


@dash_app.callback(
    Output({'type': 'manage-job-alert-div', 'index': MATCH}, 'children'),
    Output({'type': 'manage-job-buttons-div', 'index': MATCH}, 'children'),
    Output({'type': 'manage-job-status-div', 'index': MATCH}, 'children'),
    Output({'type': 'process-timestamps-div', 'index': MATCH}, 'children'),
    Input({'type': 'manage-job-button', 'action': ALL, 'index': MATCH}, 'n_clicks'),
    State('session-store', 'data'),
    prevent_initial_call = True,
)
def manage_job_button_click(
        n_clicks: Optional[int] = None,
        session_data: Optional[Dict[str, Any]] = None,
    ):
    """
    Start, stop, or pause the given job.
    """
    if not n_clicks:
        raise PreventUpdate

    session_id = session_data.get('session-id', None)
    username = active_sessions.get(session_id, {}).get('username', None)

    if not is_session_authenticated(session_id):
        success, msg = False, f"User '{username}' is not authenticated to manage jobs."
        return (
            alert_from_success_tuple((success, msg)),
            dash.no_update,
            dash.no_update,
            dash.no_update,
        )

    ctx = dash.callback_context.triggered
    if ctx[0]['value'] is None:
        raise PreventUpdate

    component_dict = json.loads(ctx[0]['prop_id'].split('.' + 'n_clicks')[0])
    daemon_id = component_dict['index']
    manage_job_action = component_dict['action']
    daemon = Daemon(daemon_id=daemon_id)

    manage_functions = {
        'start': functools.partial(daemon.run, allow_dirty_run=True),
        'stop': daemon.quit,
        'pause': daemon.pause,
    }
    if manage_job_action not in manage_functions:
        return (
            alert_from_success_tuple((False, f"Invalid action '{manage_job_action}'.")),
            dash.no_update,
            dash.no_update,
            dash.no_update,
        )

    old_status = daemon.status
    try:
        success, msg = manage_functions[manage_job_action]()
    except Exception as e:
        success, msg = False, traceback.format_exc()

    ### Wait for a status change before building the elements.
    timeout_seconds = 1.0
    check_interval_seconds = 0.01
    begin = time.perf_counter()
    while (time.perf_counter() - begin) < timeout_seconds:
        if daemon.status != old_status:
            break
        time.sleep(check_interval_seconds)

    return (
        alert_from_success_tuple((success, msg)),
        build_manage_job_buttons_div_children(daemon),
        build_status_children(daemon),
        build_process_timestamps_children(daemon),
    )


@dash_app.callback(
    Output({'type': 'manage-job-buttons-div', 'index': ALL}, 'children'),
    Output({'type': 'manage-job-status-div', 'index': ALL}, 'children'),
    Output({'type': 'process-timestamps-div', 'index': ALL}, 'children'),
    Input('refresh-jobs-interval', 'n_intervals'),
    State('session-store', 'data'),
    prevent_initial_call = True,
)
def refresh_jobs_on_interval(
        n_intervals: Optional[int] = None,
        session_data: Optional[Dict[str, Any]] = None,
    ):
    """
    When the jobs refresh interval fires, rebuild the jobs' onscreen components.
    """
    session_id = session_data.get('session-id', None)
    is_authenticated = is_session_authenticated(session_id)

    daemon_ids = [
        component_dict['id']['index']
        for component_dict in dash.callback_context.outputs_grouping[0]
    ]

    ### NOTE: The daemon may have been deleted, but the card may still exist.
    daemons = []
    for daemon_id in daemon_ids:
        try:
            daemon = Daemon(daemon_id=daemon_id)
        except Exception as e:
            daemon = None
        daemons.append(daemon)

    return (
        [
            (
                build_manage_job_buttons_div_children(daemon)
                if is_authenticated
                else []
            )
            for daemon in daemons
        ],
        [
            build_status_children(daemon)
            for daemon in daemons
        ],
        [
            build_process_timestamps_children(daemon)
            for daemon in daemons
        ],
    )
